# Copyright (C) 2005-2013, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import bz2
import os
import sys
import tempfile
from io import BytesIO

from ... import diff, errors, merge, osutils
from ... import revision as _mod_revision
from ... import tests
from ... import transport as _mod_transport
from ... import treebuilder
from ...tests import features, test_commit
from ...tree import InterTree
from .. import bzrdir, inventory, knitrepo
from ..bundle.apply_bundle import install_bundle, merge_bundle
from ..bundle.bundle_data import BundleTree
from ..bundle.serializer import read_bundle, v4, v09, write_bundle
from ..bundle.serializer.v4 import BundleSerializerV4
from ..bundle.serializer.v08 import BundleSerializerV08
from ..bundle.serializer.v09 import BundleSerializerV09
from ..inventorytree import InventoryTree


def get_text(vf, key):
    """Get the fulltext for a given revision id that is present in the vf"""
    stream = vf.get_record_stream([key], 'unordered', True)
    record = next(stream)
    return record.get_bytes_as('fulltext')


def get_inventory_text(repo, revision_id):
    """Get the fulltext for the inventory at revision id"""
    with repo.lock_read():
        return get_text(repo.inventories, (revision_id,))


class MockTree(InventoryTree):

    def __init__(self):
        from ..inventory import ROOT_ID, InventoryDirectory
        object.__init__(self)
        self.paths = {ROOT_ID: ""}
        self.ids = {"": ROOT_ID}
        self.contents = {}
        self.root = InventoryDirectory(ROOT_ID, '', None)

    inventory = property(lambda x: x)
    root_inventory = property(lambda x: x)

    def get_root_id(self):
        return self.root.file_id

    def all_file_ids(self):
        return set(self.paths.keys())

    def all_versioned_paths(self):
        return set(self.paths.values())

    def is_executable(self, path):
        # Not all the files are executable.
        return False

    def __getitem__(self, file_id):
        if file_id == self.root.file_id:
            return self.root
        else:
            return self.make_entry(file_id, self.paths[file_id])

    def get_entry_by_path(self, path):
        return self[self.path2id(path)]

    def parent_id(self, file_id):
        parent_dir = os.path.dirname(self.paths[file_id])
        if parent_dir == "":
            return None
        return self.ids[parent_dir]

    def iter_entries(self):
        for path, file_id in self.ids.items():
            yield path, self[file_id]

    def kind(self, path):
        if path in self.contents:
            kind = 'file'
        else:
            kind = 'directory'
        return kind

    def make_entry(self, file_id, path):
        from ..inventory import (InventoryDirectory, InventoryFile,
                                 InventoryLink)
        if not isinstance(file_id, bytes):
            raise TypeError(file_id)
        name = os.path.basename(path)
        kind = self.kind(path)
        parent_id = self.parent_id(file_id)
        text_sha_1, text_size = self.contents_stats(path)
        if kind == 'directory':
            ie = InventoryDirectory(file_id, name, parent_id)
        elif kind == 'file':
            ie = InventoryFile(file_id, name, parent_id)
            ie.text_sha1 = text_sha_1
            ie.text_size = text_size
        elif kind == 'symlink':
            ie = InventoryLink(file_id, name, parent_id)
        else:
            raise errors.BzrError('unknown kind %r' % kind)
        return ie

    def add_dir(self, file_id, path):
        if not isinstance(file_id, bytes):
            raise TypeError(file_id)
        self.paths[file_id] = path
        self.ids[path] = file_id

    def add_file(self, file_id, path, contents):
        if not isinstance(file_id, bytes):
            raise TypeError(file_id)
        self.add_dir(file_id, path)
        self.contents[path] = contents

    def path2id(self, path):
        return self.ids.get(path)

    def id2path(self, file_id, recurse='down'):
        try:
            return self.paths[file_id]
        except KeyError:
            raise errors.NoSuchId(file_id, self)

    def get_file(self, path):
        result = BytesIO()
        try:
            result.write(self.contents[path])
        except KeyError:
            raise _mod_transport.NoSuchFile(path)
        result.seek(0, 0)
        return result

    def get_file_revision(self, path):
        return self.inventory.get_entry_by_path(path).revision

    def get_file_size(self, path):
        return self.inventory.get_entry_by_path(path).text_size

    def get_file_sha1(self, path, file_id=None):
        return self.inventory.get_entry_by_path(path).text_sha1

    def contents_stats(self, path):
        if path not in self.contents:
            return None, None
        text_sha1 = osutils.sha_file(self.get_file(path))
        return text_sha1, len(self.contents[path])


class BTreeTester(tests.TestCase):
    """A simple unittest tester for the BundleTree class."""

    def make_tree_1(self):
        mtree = MockTree()
        mtree.add_dir(b"a", "grandparent")
        mtree.add_dir(b"b", "grandparent/parent")
        mtree.add_file(b"c", "grandparent/parent/file", b"Hello\n")
        mtree.add_dir(b"d", "grandparent/alt_parent")
        return BundleTree(mtree, b''), mtree

    def test_renames(self):
        """Ensure that file renames have the proper effect on children"""
        btree = self.make_tree_1()[0]
        self.assertEqual(btree.old_path("grandparent"), "grandparent")
        self.assertEqual(btree.old_path("grandparent/parent"),
                         "grandparent/parent")
        self.assertEqual(btree.old_path("grandparent/parent/file"),
                         "grandparent/parent/file")

        self.assertEqual(btree.id2path(b"a"), "grandparent")
        self.assertEqual(btree.id2path(b"b"), "grandparent/parent")
        self.assertEqual(btree.id2path(b"c"), "grandparent/parent/file")

        self.assertEqual(btree.path2id("grandparent"), b"a")
        self.assertEqual(btree.path2id("grandparent/parent"), b"b")
        self.assertEqual(btree.path2id("grandparent/parent/file"), b"c")

        self.assertIs(btree.path2id("grandparent2"), None)
        self.assertIs(btree.path2id("grandparent2/parent"), None)
        self.assertIs(btree.path2id("grandparent2/parent/file"), None)

        btree.note_rename("grandparent", "grandparent2")
        self.assertIs(btree.old_path("grandparent"), None)
        self.assertIs(btree.old_path("grandparent/parent"), None)
        self.assertIs(btree.old_path("grandparent/parent/file"), None)

        self.assertEqual(btree.id2path(b"a"), "grandparent2")
        self.assertEqual(btree.id2path(b"b"), "grandparent2/parent")
        self.assertEqual(btree.id2path(b"c"), "grandparent2/parent/file")

        self.assertEqual(btree.path2id("grandparent2"), b"a")
        self.assertEqual(btree.path2id("grandparent2/parent"), b"b")
        self.assertEqual(btree.path2id("grandparent2/parent/file"), b"c")

        self.assertTrue(btree.path2id("grandparent") is None)
        self.assertTrue(btree.path2id("grandparent/parent") is None)
        self.assertTrue(btree.path2id("grandparent/parent/file") is None)

        btree.note_rename("grandparent/parent", "grandparent2/parent2")
        self.assertEqual(btree.id2path(b"a"), "grandparent2")
        self.assertEqual(btree.id2path(b"b"), "grandparent2/parent2")
        self.assertEqual(btree.id2path(b"c"), "grandparent2/parent2/file")

        self.assertEqual(btree.path2id("grandparent2"), b"a")
        self.assertEqual(btree.path2id("grandparent2/parent2"), b"b")
        self.assertEqual(btree.path2id("grandparent2/parent2/file"), b"c")

        self.assertTrue(btree.path2id("grandparent2/parent") is None)
        self.assertTrue(btree.path2id("grandparent2/parent/file") is None)

        btree.note_rename("grandparent/parent/file",
                          "grandparent2/parent2/file2")
        self.assertEqual(btree.id2path(b"a"), "grandparent2")
        self.assertEqual(btree.id2path(b"b"), "grandparent2/parent2")
        self.assertEqual(btree.id2path(b"c"), "grandparent2/parent2/file2")

        self.assertEqual(btree.path2id("grandparent2"), b"a")
        self.assertEqual(btree.path2id("grandparent2/parent2"), b"b")
        self.assertEqual(btree.path2id("grandparent2/parent2/file2"), b"c")

        self.assertTrue(btree.path2id("grandparent2/parent2/file") is None)

    def test_moves(self):
        """Ensure that file moves have the proper effect on children"""
        btree = self.make_tree_1()[0]
        btree.note_rename("grandparent/parent/file",
                          "grandparent/alt_parent/file")
        self.assertEqual(btree.id2path(b"c"), "grandparent/alt_parent/file")
        self.assertEqual(btree.path2id("grandparent/alt_parent/file"), b"c")
        self.assertTrue(btree.path2id("grandparent/parent/file") is None)

    def unified_diff(self, old, new):
        out = BytesIO()
        diff.internal_diff("old", old, "new", new, out)
        out.seek(0, 0)
        return out.read()

    def make_tree_2(self):
        btree = self.make_tree_1()[0]
        btree.note_rename("grandparent/parent/file",
                          "grandparent/alt_parent/file")
        self.assertRaises(errors.NoSuchId, btree.id2path, b"e")
        self.assertFalse(btree.is_versioned("grandparent/parent/file"))
        btree.note_id(b"e", "grandparent/parent/file")
        return btree

    def test_adds(self):
        """File/inventory adds"""
        btree = self.make_tree_2()
        add_patch = self.unified_diff([], [b"Extra cheese\n"])
        btree.note_patch("grandparent/parent/file", add_patch)
        btree.note_id(b'f', 'grandparent/parent/symlink', kind='symlink')
        btree.note_target('grandparent/parent/symlink', 'venus')
        self.adds_test(btree)

    def adds_test(self, btree):
        self.assertEqual(btree.id2path(b"e"), "grandparent/parent/file")
        self.assertEqual(btree.path2id("grandparent/parent/file"), b"e")
        with btree.get_file("grandparent/parent/file") as f:
            self.assertEqual(f.read(), b"Extra cheese\n")
        self.assertEqual(
            btree.get_symlink_target('grandparent/parent/symlink'), 'venus')

    def make_tree_3(self):
        btree, mtree = self.make_tree_1()
        mtree.add_file(b"e", "grandparent/parent/topping", b"Anchovies\n")
        btree.note_rename("grandparent/parent/file",
                          "grandparent/alt_parent/file")
        btree.note_rename("grandparent/parent/topping",
                          "grandparent/alt_parent/stopping")
        return btree

    def get_file_test(self, btree):
        with btree.get_file(btree.id2path(b"e")) as f:
            self.assertEqual(f.read(), b"Lemon\n")
        with btree.get_file(btree.id2path(b"c")) as f:
            self.assertEqual(f.read(), b"Hello\n")

    def test_get_file(self):
        """Get file contents"""
        btree = self.make_tree_3()
        mod_patch = self.unified_diff([b"Anchovies\n"], [b"Lemon\n"])
        btree.note_patch("grandparent/alt_parent/stopping", mod_patch)
        self.get_file_test(btree)

    def test_delete(self):
        "Deletion by bundle"
        btree = self.make_tree_1()[0]
        with btree.get_file(btree.id2path(b"c")) as f:
            self.assertEqual(f.read(), b"Hello\n")
        btree.note_deletion("grandparent/parent/file")
        self.assertRaises(errors.NoSuchId, btree.id2path, b"c")
        self.assertFalse(btree.is_versioned("grandparent/parent/file"))

    def sorted_ids(self, tree):
        ids = sorted(tree.all_file_ids())
        return ids

    def test_iteration(self):
        """Ensure that iteration through ids works properly"""
        btree = self.make_tree_1()[0]
        self.assertEqual(self.sorted_ids(btree),
                         [inventory.ROOT_ID, b'a', b'b', b'c', b'd'])
        btree.note_deletion("grandparent/parent/file")
        btree.note_id(b"e", "grandparent/alt_parent/fool", kind="directory")
        btree.note_last_changed("grandparent/alt_parent/fool",
                                "revisionidiguess")
        self.assertEqual(self.sorted_ids(btree),
                         [inventory.ROOT_ID, b'a', b'b', b'd', b'e'])


class BundleTester1(tests.TestCaseWithTransport):

    def test_mismatched_bundle(self):
        format = bzrdir.BzrDirMetaFormat1()
        format.repository_format = knitrepo.RepositoryFormatKnit3()
        serializer = BundleSerializerV08('0.8')
        b = self.make_branch('.', format=format)
        self.assertRaises(errors.IncompatibleBundleFormat, serializer.write,
                          b.repository, [], {}, BytesIO())

    def test_matched_bundle(self):
        """Don't raise IncompatibleBundleFormat for knit2 and bundle0.9"""
        format = bzrdir.BzrDirMetaFormat1()
        format.repository_format = knitrepo.RepositoryFormatKnit3()
        serializer = BundleSerializerV09('0.9')
        b = self.make_branch('.', format=format)
        serializer.write(b.repository, [], {}, BytesIO())

    def test_mismatched_model(self):
        """Try copying a bundle from knit2 to knit1"""
        format = bzrdir.BzrDirMetaFormat1()
        format.repository_format = knitrepo.RepositoryFormatKnit3()
        source = self.make_branch_and_tree('source', format=format)
        source.commit('one', rev_id=b'one-id')
        source.commit('two', rev_id=b'two-id')
        text = BytesIO()
        write_bundle(source.branch.repository, b'two-id', b'null:', text,
                     format='0.9')
        text.seek(0)

        format = bzrdir.BzrDirMetaFormat1()
        format.repository_format = knitrepo.RepositoryFormatKnit1()
        target = self.make_branch('target', format=format)
        self.assertRaises(errors.IncompatibleRevision, install_bundle,
                          target.repository, read_bundle(text))


class BundleTester:

    def bzrdir_format(self):
        format = bzrdir.BzrDirMetaFormat1()
        format.repository_format = knitrepo.RepositoryFormatKnit1()
        return format

    def make_branch_and_tree(self, path, format=None):
        if format is None:
            format = self.bzrdir_format()
        return tests.TestCaseWithTransport.make_branch_and_tree(
            self, path, format)

    def make_branch(self, path, format=None, name=None):
        if format is None:
            format = self.bzrdir_format()
        return tests.TestCaseWithTransport.make_branch(self, path, format, name=name)

    def create_bundle_text(self, base_rev_id, rev_id):
        bundle_txt = BytesIO()
        rev_ids = write_bundle(self.b1.repository, rev_id, base_rev_id,
                               bundle_txt, format=self.format)
        bundle_txt.seek(0)
        self.assertEqual(bundle_txt.readline(),
                         b'# Bazaar revision bundle v%s\n' % self.format.encode('ascii'))
        self.assertEqual(bundle_txt.readline(), b'#\n')

        rev = self.b1.repository.get_revision(rev_id)
        self.assertEqual(bundle_txt.readline().decode('utf-8'),
                         '# message:\n')
        bundle_txt.seek(0)
        return bundle_txt, rev_ids

    def get_valid_bundle(self, base_rev_id, rev_id, checkout_dir=None):
        """Create a bundle from base_rev_id -> rev_id in built-in branch.
        Make sure that the text generated is valid, and that it
        can be applied against the base, and generate the same information.

        :return: The in-memory bundle
        """
        bundle_txt, rev_ids = self.create_bundle_text(base_rev_id, rev_id)

        # This should also validate the generated bundle
        bundle = read_bundle(bundle_txt)
        repository = self.b1.repository
        for bundle_rev in bundle.real_revisions:
            # These really should have already been checked when we read the
            # bundle, since it computes the sha1 hash for the revision, which
            # only will match if everything is okay, but lets be explicit about
            # it
            branch_rev = repository.get_revision(bundle_rev.revision_id)
            for a in ('inventory_sha1', 'revision_id', 'parent_ids',
                      'timestamp', 'timezone', 'message', 'committer',
                      'parent_ids', 'properties'):
                self.assertEqual(getattr(branch_rev, a),
                                 getattr(bundle_rev, a))
            self.assertEqual(len(branch_rev.parent_ids),
                             len(bundle_rev.parent_ids))
        self.assertEqual(rev_ids,
                         [r.revision_id for r in bundle.real_revisions])
        self.valid_apply_bundle(base_rev_id, bundle, checkout_dir=checkout_dir)

        return bundle

    def get_invalid_bundle(self, base_rev_id, rev_id):
        """Create a bundle from base_rev_id -> rev_id in built-in branch.
        Munge the text so that it's invalid.

        :return: The in-memory bundle
        """
        bundle_txt, rev_ids = self.create_bundle_text(base_rev_id, rev_id)
        new_text = bundle_txt.getvalue().replace(b'executable:no',
                                                 b'executable:yes')
        bundle_txt = BytesIO(new_text)
        bundle = read_bundle(bundle_txt)
        self.valid_apply_bundle(base_rev_id, bundle)
        return bundle

    def test_non_bundle(self):
        self.assertRaises(errors.NotABundle,
                          read_bundle, BytesIO(b'#!/bin/sh\n'))

    def test_malformed(self):
        self.assertRaises(errors.BadBundle, read_bundle,
                          BytesIO(b'# Bazaar revision bundle v'))

    def test_crlf_bundle(self):
        try:
            read_bundle(BytesIO(b'# Bazaar revision bundle v0.8\r\n'))
        except errors.BadBundle:
            # It is currently permitted for bundles with crlf line endings to
            # make read_bundle raise a BadBundle, but this should be fixed.
            # Anything else, especially NotABundle, is an error.
            pass

    def get_checkout(self, rev_id, checkout_dir=None):
        """Get a new tree, with the specified revision in it.
        """

        if checkout_dir is None:
            checkout_dir = tempfile.mkdtemp(prefix='test-branch-', dir='.')
        else:
            if not os.path.exists(checkout_dir):
                os.mkdir(checkout_dir)
        tree = self.make_branch_and_tree(checkout_dir)
        s = BytesIO()
        ancestors = write_bundle(self.b1.repository, rev_id, b'null:', s,
                                 format=self.format)
        s.seek(0)
        self.assertIsInstance(s.getvalue(), bytes)
        install_bundle(tree.branch.repository, read_bundle(s))
        for ancestor in ancestors:
            old = self.b1.repository.revision_tree(ancestor)
            new = tree.branch.repository.revision_tree(ancestor)
            with old.lock_read(), new.lock_read():
                # Check that there aren't any inventory level changes
                delta = new.changes_from(old)
                self.assertFalse(delta.has_changed(),
                                 'Revision %s not copied correctly.'
                                 % (ancestor,))

                # Now check that the file contents are all correct
                for path in old.all_versioned_paths():
                    try:
                        old_file = old.get_file(path)
                    except _mod_transport.NoSuchFile:
                        continue
                    self.assertEqual(
                        old_file.read(), new.get_file(path).read())
        if not _mod_revision.is_null(rev_id):
            tree.branch.generate_revision_history(rev_id)
            tree.update()
            delta = tree.changes_from(self.b1.repository.revision_tree(rev_id))
            self.assertFalse(delta.has_changed(),
                             'Working tree has modifications: %s' % delta)
        return tree

    def valid_apply_bundle(self, base_rev_id, info, checkout_dir=None):
        """Get the base revision, apply the changes, and make
        sure everything matches the builtin branch.
        """
        to_tree = self.get_checkout(base_rev_id, checkout_dir=checkout_dir)
        to_tree.lock_write()
        try:
            self._valid_apply_bundle(base_rev_id, info, to_tree)
        finally:
            to_tree.unlock()

    def _valid_apply_bundle(self, base_rev_id, info, to_tree):
        original_parents = to_tree.get_parent_ids()
        repository = to_tree.branch.repository
        original_parents = to_tree.get_parent_ids()
        self.assertIs(repository.has_revision(base_rev_id), True)
        for rev in info.real_revisions:
            self.assertTrue(not repository.has_revision(rev.revision_id),
                            'Revision {%s} present before applying bundle'
                            % rev.revision_id)
        merge_bundle(info, to_tree, True, merge.Merge3Merger, False, False)

        for rev in info.real_revisions:
            self.assertTrue(repository.has_revision(rev.revision_id),
                            'Missing revision {%s} after applying bundle'
                            % rev.revision_id)

        self.assertTrue(to_tree.branch.repository.has_revision(info.target))
        # Do we also want to verify that all the texts have been added?

        self.assertEqual(original_parents + [info.target],
                         to_tree.get_parent_ids())

        rev = info.real_revisions[-1]
        base_tree = self.b1.repository.revision_tree(rev.revision_id)
        to_tree = to_tree.branch.repository.revision_tree(rev.revision_id)

        # TODO: make sure the target tree is identical to base tree
        #       we might also check the working tree.

        base_files = list(base_tree.list_files())
        to_files = list(to_tree.list_files())
        self.assertEqual(len(base_files), len(to_files))
        for base_file, to_file in zip(base_files, to_files):
            self.assertEqual(base_file, to_file)

        for path, status, kind, entry in base_files:
            # Check that the meta information is the same
            to_path = InterTree.get(base_tree, to_tree).find_target_path(path)
            self.assertEqual(
                base_tree.get_file_size(path),
                to_tree.get_file_size(to_path))
            self.assertEqual(
                base_tree.get_file_sha1(path),
                to_tree.get_file_sha1(to_path))
            # Check that the contents are the same
            # This is pretty expensive
            # self.assertEqual(base_tree.get_file(fileid).read(),
            #         to_tree.get_file(fileid).read())

    def test_bundle(self):
        self.tree1 = self.make_branch_and_tree('b1')
        self.b1 = self.tree1.branch

        self.build_tree_contents([('b1/one', b'one\n')])
        self.tree1.add('one', ids=b'one-id')
        self.tree1.set_root_id(b'root-id')
        self.tree1.commit('add one', rev_id=b'a@cset-0-1')

        bundle = self.get_valid_bundle(b'null:', b'a@cset-0-1')

        # Make sure we can handle files with spaces, tabs, other
        # bogus characters
        self.build_tree([
            'b1/with space.txt', 'b1/dir/', 'b1/dir/filein subdir.c', 'b1/dir/WithCaps.txt', 'b1/dir/ pre space', 'b1/sub/', 'b1/sub/sub/', 'b1/sub/sub/nonempty.txt'
            ])
        self.build_tree_contents([('b1/sub/sub/emptyfile.txt', b''),
                                  ('b1/dir/nolastnewline.txt', b'bloop')])
        tt = self.tree1.transform()
        tt.new_file('executable', tt.root, [b'#!/bin/sh\n'], b'exe-1', True)
        tt.apply()
        # have to fix length of file-id so that we can predictably rewrite
        # a (length-prefixed) record containing it later.
        self.tree1.add('with space.txt', ids=b'withspace-id')
        self.tree1.add([
            'dir', 'dir/filein subdir.c', 'dir/WithCaps.txt', 'dir/ pre space', 'dir/nolastnewline.txt', 'sub', 'sub/sub', 'sub/sub/nonempty.txt', 'sub/sub/emptyfile.txt'
            ])
        self.tree1.commit('add whitespace', rev_id=b'a@cset-0-2')

        bundle = self.get_valid_bundle(b'a@cset-0-1', b'a@cset-0-2')

        # Check a rollup bundle
        bundle = self.get_valid_bundle(b'null:', b'a@cset-0-2')

        # Now delete entries
        self.tree1.remove(
            ['sub/sub/nonempty.txt', 'sub/sub/emptyfile.txt', 'sub/sub'
             ])
        tt = self.tree1.transform()
        trans_id = tt.trans_id_tree_path('executable')
        tt.set_executability(False, trans_id)
        tt.apply()
        self.tree1.commit('removed', rev_id=b'a@cset-0-3')

        bundle = self.get_valid_bundle(b'a@cset-0-2', b'a@cset-0-3')
        self.assertRaises((errors.TestamentMismatch,
                           errors.VersionedFileInvalidChecksum,
                           errors.BadBundle), self.get_invalid_bundle,
                          b'a@cset-0-2', b'a@cset-0-3')
        # Check a rollup bundle
        bundle = self.get_valid_bundle(b'null:', b'a@cset-0-3')

        # Now move the directory
        self.tree1.rename_one('dir', 'sub/dir')
        self.tree1.commit('rename dir', rev_id=b'a@cset-0-4')

        bundle = self.get_valid_bundle(b'a@cset-0-3', b'a@cset-0-4')
        # Check a rollup bundle
        bundle = self.get_valid_bundle(b'null:', b'a@cset-0-4')

        # Modified files
        with open('b1/sub/dir/WithCaps.txt', 'ab') as f:
            f.write(b'\nAdding some text\n')
        with open('b1/sub/dir/ pre space', 'ab') as f:
            f.write(
                b'\r\nAdding some\r\nDOS format lines\r\n')
        with open('b1/sub/dir/nolastnewline.txt', 'ab') as f:
            f.write(b'\n')
        self.tree1.rename_one('sub/dir/ pre space',
                              'sub/ start space')
        self.tree1.commit('Modified files', rev_id=b'a@cset-0-5')
        bundle = self.get_valid_bundle(b'a@cset-0-4', b'a@cset-0-5')

        self.tree1.rename_one('sub/dir/WithCaps.txt', 'temp')
        self.tree1.rename_one('with space.txt', 'WithCaps.txt')
        self.tree1.rename_one('temp', 'with space.txt')
        self.tree1.commit('swap filenames', rev_id=b'a@cset-0-6',
                          verbose=False)
        bundle = self.get_valid_bundle(b'a@cset-0-5', b'a@cset-0-6')
        other = self.get_checkout(b'a@cset-0-5')
        tree1_inv = get_inventory_text(self.tree1.branch.repository,
                                       b'a@cset-0-5')
        tree2_inv = get_inventory_text(other.branch.repository,
                                       b'a@cset-0-5')
        self.assertEqualDiff(tree1_inv, tree2_inv)
        other.rename_one('sub/dir/nolastnewline.txt', 'sub/nolastnewline.txt')
        other.commit('rename file', rev_id=b'a@cset-0-6b')
        self.tree1.merge_from_branch(other.branch)
        self.tree1.commit('Merge', rev_id=b'a@cset-0-7',
                          verbose=False)
        bundle = self.get_valid_bundle(b'a@cset-0-6', b'a@cset-0-7')

    def _test_symlink_bundle(self, link_name, link_target, new_link_target):
        link_id = b'link-1'

        self.requireFeature(features.SymlinkFeature(self.test_dir))
        self.tree1 = self.make_branch_and_tree('b1')
        self.b1 = self.tree1.branch

        tt = self.tree1.transform()
        tt.new_symlink(link_name, tt.root, link_target, link_id)
        tt.apply()
        self.tree1.commit('add symlink', rev_id=b'l@cset-0-1')
        bundle = self.get_valid_bundle(b'null:', b'l@cset-0-1')
        if getattr(bundle, 'revision_tree', None) is not None:
            # Not all bundle formats supports revision_tree
            bund_tree = bundle.revision_tree(self.b1.repository, b'l@cset-0-1')
            self.assertEqual(
                link_target, bund_tree.get_symlink_target(link_name))

        tt = self.tree1.transform()
        trans_id = tt.trans_id_tree_path(link_name)
        tt.adjust_path('link2', tt.root, trans_id)
        tt.delete_contents(trans_id)
        tt.create_symlink(new_link_target, trans_id)
        tt.apply()
        self.tree1.commit('rename and change symlink', rev_id=b'l@cset-0-2')
        bundle = self.get_valid_bundle(b'l@cset-0-1', b'l@cset-0-2')
        if getattr(bundle, 'revision_tree', None) is not None:
            # Not all bundle formats supports revision_tree
            bund_tree = bundle.revision_tree(self.b1.repository, b'l@cset-0-2')
            self.assertEqual(new_link_target,
                             bund_tree.get_symlink_target('link2'))

        tt = self.tree1.transform()
        trans_id = tt.trans_id_tree_path('link2')
        tt.delete_contents(trans_id)
        tt.create_symlink('jupiter', trans_id)
        tt.apply()
        self.tree1.commit('just change symlink target', rev_id=b'l@cset-0-3')
        bundle = self.get_valid_bundle(b'l@cset-0-2', b'l@cset-0-3')

        tt = self.tree1.transform()
        trans_id = tt.trans_id_tree_path('link2')
        tt.delete_contents(trans_id)
        tt.apply()
        self.tree1.commit('Delete symlink', rev_id=b'l@cset-0-4')
        bundle = self.get_valid_bundle(b'l@cset-0-3', b'l@cset-0-4')

    def test_symlink_bundle(self):
        self._test_symlink_bundle('link', 'bar/foo', 'mars')

    def test_unicode_symlink_bundle(self):
        self.requireFeature(features.UnicodeFilenameFeature)
        self._test_symlink_bundle('\N{Euro Sign}link',
                                  'bar/\N{Euro Sign}foo',
                                  'mars\N{Euro Sign}')

    def test_binary_bundle(self):
        self.tree1 = self.make_branch_and_tree('b1')
        self.b1 = self.tree1.branch
        tt = self.tree1.transform()

        # Add
        tt.new_file('file', tt.root, [
                    b'\x00\n\x00\r\x01\n\x02\r\xff'], b'binary-1')
        tt.new_file('file2', tt.root, [b'\x01\n\x02\r\x03\n\x04\r\xff'],
                    b'binary-2')
        tt.apply()
        self.tree1.commit('add binary', rev_id=b'b@cset-0-1')
        self.get_valid_bundle(b'null:', b'b@cset-0-1')

        # Delete
        tt = self.tree1.transform()
        trans_id = tt.trans_id_tree_path('file')
        tt.delete_contents(trans_id)
        tt.apply()
        self.tree1.commit('delete binary', rev_id=b'b@cset-0-2')
        self.get_valid_bundle(b'b@cset-0-1', b'b@cset-0-2')

        # Rename & modify
        tt = self.tree1.transform()
        trans_id = tt.trans_id_tree_path('file2')
        tt.adjust_path('file3', tt.root, trans_id)
        tt.delete_contents(trans_id)
        tt.create_file([b'file\rcontents\x00\n\x00'], trans_id)
        tt.apply()
        self.tree1.commit('rename and modify binary', rev_id=b'b@cset-0-3')
        self.get_valid_bundle(b'b@cset-0-2', b'b@cset-0-3')

        # Modify
        tt = self.tree1.transform()
        trans_id = tt.trans_id_tree_path('file3')
        tt.delete_contents(trans_id)
        tt.create_file([b'\x00file\rcontents'], trans_id)
        tt.apply()
        self.tree1.commit('just modify binary', rev_id=b'b@cset-0-4')
        self.get_valid_bundle(b'b@cset-0-3', b'b@cset-0-4')

        # Rollup
        self.get_valid_bundle(b'null:', b'b@cset-0-4')

    def test_last_modified(self):
        self.tree1 = self.make_branch_and_tree('b1')
        self.b1 = self.tree1.branch
        tt = self.tree1.transform()
        tt.new_file('file', tt.root, [b'file'], b'file')
        tt.apply()
        self.tree1.commit('create file', rev_id=b'a@lmod-0-1')

        tt = self.tree1.transform()
        trans_id = tt.trans_id_tree_path('file')
        tt.delete_contents(trans_id)
        tt.create_file([b'file2'], trans_id)
        tt.apply()
        self.tree1.commit('modify text', rev_id=b'a@lmod-0-2a')

        other = self.get_checkout(b'a@lmod-0-1')
        tt = other.transform()
        trans_id = tt.trans_id_tree_path('file2')
        tt.delete_contents(trans_id)
        tt.create_file([b'file2'], trans_id)
        tt.apply()
        other.commit('modify text in another tree', rev_id=b'a@lmod-0-2b')
        self.tree1.merge_from_branch(other.branch)
        self.tree1.commit('Merge', rev_id=b'a@lmod-0-3',
                          verbose=False)
        self.tree1.commit('Merge', rev_id=b'a@lmod-0-4')
        bundle = self.get_valid_bundle(b'a@lmod-0-2a', b'a@lmod-0-4')

    def test_hide_history(self):
        self.tree1 = self.make_branch_and_tree('b1')
        self.b1 = self.tree1.branch

        with open('b1/one', 'wb') as f:
            f.write(b'one\n')
        self.tree1.add('one')
        self.tree1.commit('add file', rev_id=b'a@cset-0-1')
        with open('b1/one', 'wb') as f:
            f.write(b'two\n')
        self.tree1.commit('modify', rev_id=b'a@cset-0-2')
        with open('b1/one', 'wb') as f:
            f.write(b'three\n')
        self.tree1.commit('modify', rev_id=b'a@cset-0-3')
        bundle_file = BytesIO()
        rev_ids = write_bundle(self.tree1.branch.repository, b'a@cset-0-3',
                               b'a@cset-0-1', bundle_file, format=self.format)
        self.assertNotContainsRe(bundle_file.getvalue(), b'\btwo\b')
        self.assertContainsRe(self.get_raw(bundle_file), b'one')
        self.assertContainsRe(self.get_raw(bundle_file), b'three')

    def test_bundle_same_basis(self):
        """Ensure using the basis as the target doesn't cause an error"""
        self.tree1 = self.make_branch_and_tree('b1')
        self.tree1.commit('add file', rev_id=b'a@cset-0-1')
        bundle_file = BytesIO()
        rev_ids = write_bundle(self.tree1.branch.repository, b'a@cset-0-1',
                               b'a@cset-0-1', bundle_file)

    @staticmethod
    def get_raw(bundle_file):
        return bundle_file.getvalue()

    def test_unicode_bundle(self):
        self.requireFeature(features.UnicodeFilenameFeature)
        # Handle international characters
        os.mkdir('b1')
        f = open('b1/with Dod\N{Euro Sign}', 'wb')

        self.tree1 = self.make_branch_and_tree('b1')
        self.b1 = self.tree1.branch

        f.write(('A file\n'
                 'With international man of mystery\n'
                 'William Dod\xe9\n').encode())
        f.close()

        self.tree1.add(['with Dod\N{Euro Sign}'], ids=[b'withdod-id'])
        self.tree1.commit('i18n commit from William Dod\xe9',
                          rev_id=b'i18n-1', committer='William Dod\xe9')

        # Add
        bundle = self.get_valid_bundle(b'null:', b'i18n-1')

        # Modified
        f = open('b1/with Dod\N{Euro Sign}', 'wb')
        f.write('Modified \xb5\n'.encode())
        f.close()
        self.tree1.commit('modified', rev_id=b'i18n-2')

        bundle = self.get_valid_bundle(b'i18n-1', b'i18n-2')

        # Renamed
        self.tree1.rename_one('with Dod\N{Euro Sign}', 'B\N{Euro Sign}gfors')
        self.tree1.commit('renamed, the new i18n man', rev_id=b'i18n-3',
                          committer='Erik B\xe5gfors')

        bundle = self.get_valid_bundle(b'i18n-2', b'i18n-3')

        # Removed
        self.tree1.remove(['B\N{Euro Sign}gfors'])
        self.tree1.commit('removed', rev_id=b'i18n-4')

        bundle = self.get_valid_bundle(b'i18n-3', b'i18n-4')

        # Rollup
        bundle = self.get_valid_bundle(b'null:', b'i18n-4')

    def test_whitespace_bundle(self):
        if sys.platform in ('win32', 'cygwin'):
            raise tests.TestSkipped('Windows doesn\'t support filenames'
                                    ' with tabs or trailing spaces')
        self.tree1 = self.make_branch_and_tree('b1')
        self.b1 = self.tree1.branch

        self.build_tree(['b1/trailing space '])
        self.tree1.add(['trailing space '])
        # TODO: jam 20060701 Check for handling files with '\t' characters
        #       once we actually support them

        # Added
        self.tree1.commit('funky whitespace', rev_id=b'white-1')

        bundle = self.get_valid_bundle(b'null:', b'white-1')

        # Modified
        with open('b1/trailing space ', 'ab') as f:
            f.write(b'add some text\n')
        self.tree1.commit('add text', rev_id=b'white-2')

        bundle = self.get_valid_bundle(b'white-1', b'white-2')

        # Renamed
        self.tree1.rename_one('trailing space ', ' start and end space ')
        self.tree1.commit('rename', rev_id=b'white-3')

        bundle = self.get_valid_bundle(b'white-2', b'white-3')

        # Removed
        self.tree1.remove([' start and end space '])
        self.tree1.commit('removed', rev_id=b'white-4')

        bundle = self.get_valid_bundle(b'white-3', b'white-4')

        # Now test a complet roll-up
        bundle = self.get_valid_bundle(b'null:', b'white-4')

    def test_alt_timezone_bundle(self):
        self.tree1 = self.make_branch_and_memory_tree('b1')
        self.b1 = self.tree1.branch
        builder = treebuilder.TreeBuilder()

        self.tree1.lock_write()
        builder.start_tree(self.tree1)
        builder.build(['newfile'])
        builder.finish_tree()

        # Asia/Colombo offset = 5 hours 30 minutes
        self.tree1.commit('non-hour offset timezone', rev_id=b'tz-1',
                          timezone=19800, timestamp=1152544886.0)

        bundle = self.get_valid_bundle(b'null:', b'tz-1')

        rev = bundle.revisions[0]
        self.assertEqual('Mon 2006-07-10 20:51:26.000000000 +0530', rev.date)
        self.assertEqual(19800, rev.timezone)
        self.assertEqual(1152544886.0, rev.timestamp)
        self.tree1.unlock()

    def test_bundle_root_id(self):
        self.tree1 = self.make_branch_and_tree('b1')
        self.b1 = self.tree1.branch
        self.tree1.commit('message', rev_id=b'revid1')
        bundle = self.get_valid_bundle(b'null:', b'revid1')
        tree = self.get_bundle_tree(bundle, b'revid1')
        root_revision = tree.get_file_revision('')
        self.assertEqual(b'revid1', root_revision)

    def test_install_revisions(self):
        self.tree1 = self.make_branch_and_tree('b1')
        self.b1 = self.tree1.branch
        self.tree1.commit('message', rev_id=b'rev2a')
        bundle = self.get_valid_bundle(b'null:', b'rev2a')
        branch2 = self.make_branch('b2')
        self.assertFalse(branch2.repository.has_revision(b'rev2a'))
        target_revision = bundle.install_revisions(branch2.repository)
        self.assertTrue(branch2.repository.has_revision(b'rev2a'))
        self.assertEqual(b'rev2a', target_revision)

    def test_bundle_empty_property(self):
        """Test serializing revision properties with an empty value."""
        tree = self.make_branch_and_memory_tree('tree')
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.add([''], ids=[b'TREE_ROOT'])
        tree.commit('One', revprops={'one': 'two',
                                     'empty': ''}, rev_id=b'rev1')
        self.b1 = tree.branch
        bundle_sio, revision_ids = self.create_bundle_text(b'null:', b'rev1')
        bundle = read_bundle(bundle_sio)
        revision_info = bundle.revisions[0]
        self.assertEqual(b'rev1', revision_info.revision_id)
        rev = revision_info.as_revision()
        self.assertEqual({'branch-nick': 'tree', 'empty': '', 'one': 'two'},
                         rev.properties)

    def test_bundle_sorted_properties(self):
        """For stability the writer should write properties in sorted order."""
        tree = self.make_branch_and_memory_tree('tree')
        tree.lock_write()
        self.addCleanup(tree.unlock)

        tree.add([''], ids=[b'TREE_ROOT'])
        tree.commit('One', rev_id=b'rev1',
                    revprops={'a': '4', 'b': '3', 'c': '2', 'd': '1'})
        self.b1 = tree.branch
        bundle_sio, revision_ids = self.create_bundle_text(b'null:', b'rev1')
        bundle = read_bundle(bundle_sio)
        revision_info = bundle.revisions[0]
        self.assertEqual(b'rev1', revision_info.revision_id)
        rev = revision_info.as_revision()
        self.assertEqual({'branch-nick': 'tree', 'a': '4', 'b': '3', 'c': '2',
                          'd': '1'}, rev.properties)

    def test_bundle_unicode_properties(self):
        """We should be able to round trip a non-ascii property."""
        tree = self.make_branch_and_memory_tree('tree')
        tree.lock_write()
        self.addCleanup(tree.unlock)

        tree.add([''], ids=[b'TREE_ROOT'])
        # Revisions themselves do not require anything about revision property
        # keys, other than that they are a basestring, and do not contain
        # whitespace.
        # However, Testaments assert than they are str(), and thus should not
        # be Unicode.
        tree.commit('One', rev_id=b'rev1',
                    revprops={'omega': '\u03a9', 'alpha': '\u03b1'})
        self.b1 = tree.branch
        bundle_sio, revision_ids = self.create_bundle_text(b'null:', b'rev1')
        bundle = read_bundle(bundle_sio)
        revision_info = bundle.revisions[0]
        self.assertEqual(b'rev1', revision_info.revision_id)
        rev = revision_info.as_revision()
        self.assertEqual({'branch-nick': 'tree', 'omega': '\u03a9',
                          'alpha': '\u03b1'}, rev.properties)

    def test_bundle_with_ghosts(self):
        tree = self.make_branch_and_tree('tree')
        self.b1 = tree.branch
        self.build_tree_contents([('tree/file', b'content1')])
        tree.add(['file'])
        tree.commit('rev1')
        self.build_tree_contents([('tree/file', b'content2')])
        tree.add_parent_tree_id(b'ghost')
        tree.commit('rev2', rev_id=b'rev2')
        bundle = self.get_valid_bundle(b'null:', b'rev2')

    def make_simple_tree(self, format=None):
        tree = self.make_branch_and_tree('b1', format=format)
        self.b1 = tree.branch
        self.build_tree(['b1/file'])
        tree.add('file')
        return tree

    def test_across_serializers(self):
        tree = self.make_simple_tree('knit')
        tree.commit('hello', rev_id=b'rev1')
        tree.commit('hello', rev_id=b'rev2')
        bundle = read_bundle(self.create_bundle_text(b'null:', b'rev2')[0])
        repo = self.make_repository('repo', format='dirstate-with-subtree')
        bundle.install_revisions(repo)
        inv_text = b''.join(repo._get_inventory_xml(b'rev2'))
        self.assertNotContainsRe(inv_text, b'format="5"')
        self.assertContainsRe(inv_text, b'format="7"')

    def make_repo_with_installed_revisions(self):
        tree = self.make_simple_tree('knit')
        tree.commit('hello', rev_id=b'rev1')
        tree.commit('hello', rev_id=b'rev2')
        bundle = read_bundle(self.create_bundle_text(b'null:', b'rev2')[0])
        repo = self.make_repository('repo', format='dirstate-with-subtree')
        bundle.install_revisions(repo)
        return repo

    def test_across_models(self):
        repo = self.make_repo_with_installed_revisions()
        inv = repo.get_inventory(b'rev2')
        self.assertEqual(b'rev2', inv.root.revision)
        root_id = inv.root.file_id
        repo.lock_read()
        self.addCleanup(repo.unlock)
        self.assertEqual({(root_id, b'rev1'): (),
                          (root_id, b'rev2'): ((root_id, b'rev1'),)},
                         repo.texts.get_parent_map([(root_id, b'rev1'), (root_id, b'rev2')]))

    def test_inv_hash_across_serializers(self):
        repo = self.make_repo_with_installed_revisions()
        recorded_inv_sha1 = repo.get_revision(b'rev2').inventory_sha1
        xml = b''.join(repo._get_inventory_xml(b'rev2'))
        self.assertEqual(osutils.sha_string(xml), recorded_inv_sha1)

    def test_across_models_incompatible(self):
        tree = self.make_simple_tree('dirstate-with-subtree')
        tree.commit('hello', rev_id=b'rev1')
        tree.commit('hello', rev_id=b'rev2')
        try:
            bundle = read_bundle(self.create_bundle_text(b'null:', b'rev1')[0])
        except errors.IncompatibleBundleFormat:
            raise tests.TestSkipped("Format 0.8 doesn't work with knit3")
        repo = self.make_repository('repo', format='knit')
        bundle.install_revisions(repo)

        bundle = read_bundle(self.create_bundle_text(b'null:', b'rev2')[0])
        self.assertRaises(errors.IncompatibleRevision,
                          bundle.install_revisions, repo)

    def test_get_merge_request(self):
        tree = self.make_simple_tree()
        tree.commit('hello', rev_id=b'rev1')
        tree.commit('hello', rev_id=b'rev2')
        bundle = read_bundle(self.create_bundle_text(b'null:', b'rev1')[0])
        result = bundle.get_merge_request(tree.branch.repository)
        self.assertEqual((None, b'rev1', 'inapplicable'), result)

    def test_with_subtree(self):
        tree = self.make_branch_and_tree('tree',
                                         format='dirstate-with-subtree')
        self.b1 = tree.branch
        subtree = self.make_branch_and_tree('tree/subtree',
                                            format='dirstate-with-subtree')
        tree.add('subtree')
        tree.commit('hello', rev_id=b'rev1')
        try:
            bundle = read_bundle(self.create_bundle_text(b'null:', b'rev1')[0])
        except errors.IncompatibleBundleFormat:
            raise tests.TestSkipped("Format 0.8 doesn't work with knit3")
        if isinstance(bundle, v09.BundleInfo09):
            raise tests.TestSkipped("Format 0.9 doesn't work with subtrees")
        repo = self.make_repository('repo', format='knit')
        self.assertRaises(errors.IncompatibleRevision,
                          bundle.install_revisions, repo)
        repo2 = self.make_repository('repo2', format='dirstate-with-subtree')
        bundle.install_revisions(repo2)

    def test_revision_id_with_slash(self):
        self.tree1 = self.make_branch_and_tree('tree')
        self.b1 = self.tree1.branch
        try:
            self.tree1.commit('Revision/id/with/slashes', rev_id=b'rev/id')
        except ValueError:
            raise tests.TestSkipped(
                "Repository doesn't support revision ids with slashes")
        bundle = self.get_valid_bundle(b'null:', b'rev/id')

    def test_skip_file(self):
        """Make sure we don't accidentally write to the wrong versionedfile"""
        self.tree1 = self.make_branch_and_tree('tree')
        self.b1 = self.tree1.branch
        # rev1 is not present in bundle, done by fetch
        self.build_tree_contents([('tree/file2', b'contents1')])
        self.tree1.add('file2', ids=b'file2-id')
        self.tree1.commit('rev1', rev_id=b'reva')
        self.build_tree_contents([('tree/file3', b'contents2')])
        # rev2 is present in bundle, and done by fetch
        # having file1 in the bunle causes file1's versionedfile to be opened.
        self.tree1.add('file3', ids=b'file3-id')
        rev2 = self.tree1.commit('rev2')
        # Updating file2 should not cause an attempt to add to file1's vf
        target = self.tree1.controldir.sprout('target').open_workingtree()
        self.build_tree_contents([('tree/file2', b'contents3')])
        self.tree1.commit('rev3', rev_id=b'rev3')
        bundle = self.get_valid_bundle(b'reva', b'rev3')
        if getattr(bundle, 'get_bundle_reader', None) is None:
            raise tests.TestSkipped('Bundle format cannot provide reader')
        file_ids = {
            (f, r) for b, m, k, r, f in bundle.get_bundle_reader().iter_records()
            if f is not None}
        self.assertEqual(
            {(b'file2-id', b'rev3'), (b'file3-id', rev2)}, file_ids)
        bundle.install_revisions(target.branch.repository)


class V08BundleTester(BundleTester, tests.TestCaseWithTransport):

    format = '0.8'

    def test_bundle_empty_property(self):
        """Test serializing revision properties with an empty value."""
        tree = self.make_branch_and_memory_tree('tree')
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.add([''], ids=[b'TREE_ROOT'])
        tree.commit('One', revprops={'one': 'two',
                                     'empty': ''}, rev_id=b'rev1')
        self.b1 = tree.branch
        bundle_sio, revision_ids = self.create_bundle_text(b'null:', b'rev1')
        self.assertContainsRe(bundle_sio.getvalue(),
                              b'# properties:\n'
                              b'#   branch-nick: tree\n'
                              b'#   empty: \n'
                              b'#   one: two\n'
                              )
        bundle = read_bundle(bundle_sio)
        revision_info = bundle.revisions[0]
        self.assertEqual(b'rev1', revision_info.revision_id)
        rev = revision_info.as_revision()
        self.assertEqual({'branch-nick': 'tree', 'empty': '', 'one': 'two'},
                         rev.properties)

    def get_bundle_tree(self, bundle, revision_id):
        repository = self.make_repository('repo')
        return bundle.revision_tree(repository, b'revid1')

    def test_bundle_empty_property_alt(self):
        """Test serializing revision properties with an empty value.

        Older readers had a bug when reading an empty property.
        They assumed that all keys ended in ': \n'. However they would write an
        empty value as ':\n'. This tests make sure that all newer bzr versions
        can handle th second form.
        """
        tree = self.make_branch_and_memory_tree('tree')
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.add([''], ids=[b'TREE_ROOT'])
        tree.commit('One', revprops={'one': 'two',
                                     'empty': ''}, rev_id=b'rev1')
        self.b1 = tree.branch
        bundle_sio, revision_ids = self.create_bundle_text(b'null:', b'rev1')
        txt = bundle_sio.getvalue()
        loc = txt.find(b'#   empty: ') + len(b'#   empty:')
        # Create a new bundle, which strips the trailing space after empty
        bundle_sio = BytesIO(txt[:loc] + txt[loc + 1:])

        self.assertContainsRe(bundle_sio.getvalue(),
                              b'# properties:\n'
                              b'#   branch-nick: tree\n'
                              b'#   empty:\n'
                              b'#   one: two\n'
                              )
        bundle = read_bundle(bundle_sio)
        revision_info = bundle.revisions[0]
        self.assertEqual(b'rev1', revision_info.revision_id)
        rev = revision_info.as_revision()
        self.assertEqual({'branch-nick': 'tree', 'empty': '', 'one': 'two'},
                         rev.properties)

    def test_bundle_sorted_properties(self):
        """For stability the writer should write properties in sorted order."""
        tree = self.make_branch_and_memory_tree('tree')
        tree.lock_write()
        self.addCleanup(tree.unlock)

        tree.add([''], ids=[b'TREE_ROOT'])
        tree.commit('One', rev_id=b'rev1',
                    revprops={'a': '4', 'b': '3', 'c': '2', 'd': '1'})
        self.b1 = tree.branch
        bundle_sio, revision_ids = self.create_bundle_text(b'null:', b'rev1')
        self.assertContainsRe(bundle_sio.getvalue(),
                              b'# properties:\n'
                              b'#   a: 4\n'
                              b'#   b: 3\n'
                              b'#   branch-nick: tree\n'
                              b'#   c: 2\n'
                              b'#   d: 1\n'
                              )
        bundle = read_bundle(bundle_sio)
        revision_info = bundle.revisions[0]
        self.assertEqual(b'rev1', revision_info.revision_id)
        rev = revision_info.as_revision()
        self.assertEqual({'branch-nick': 'tree', 'a': '4', 'b': '3', 'c': '2',
                          'd': '1'}, rev.properties)

    def test_bundle_unicode_properties(self):
        """We should be able to round trip a non-ascii property."""
        tree = self.make_branch_and_memory_tree('tree')
        tree.lock_write()
        self.addCleanup(tree.unlock)

        tree.add([''], ids=[b'TREE_ROOT'])
        # Revisions themselves do not require anything about revision property
        # keys, other than that they are a basestring, and do not contain
        # whitespace.
        # However, Testaments assert than they are str(), and thus should not
        # be Unicode.
        tree.commit('One', rev_id=b'rev1',
                    revprops={'omega': '\u03a9', 'alpha': '\u03b1'})
        self.b1 = tree.branch
        bundle_sio, revision_ids = self.create_bundle_text(b'null:', b'rev1')
        self.assertContainsRe(bundle_sio.getvalue(),
                              b'# properties:\n'
                              b'#   alpha: \xce\xb1\n'
                              b'#   branch-nick: tree\n'
                              b'#   omega: \xce\xa9\n'
                              )
        bundle = read_bundle(bundle_sio)
        revision_info = bundle.revisions[0]
        self.assertEqual(b'rev1', revision_info.revision_id)
        rev = revision_info.as_revision()
        self.assertEqual({'branch-nick': 'tree', 'omega': '\u03a9',
                          'alpha': '\u03b1'}, rev.properties)


class V09BundleKnit2Tester(V08BundleTester):

    format = '0.9'

    def bzrdir_format(self):
        format = bzrdir.BzrDirMetaFormat1()
        format.repository_format = knitrepo.RepositoryFormatKnit3()
        return format


class V09BundleKnit1Tester(V08BundleTester):

    format = '0.9'

    def bzrdir_format(self):
        format = bzrdir.BzrDirMetaFormat1()
        format.repository_format = knitrepo.RepositoryFormatKnit1()
        return format


class V4BundleTester(BundleTester, tests.TestCaseWithTransport):

    format = '4'

    def get_valid_bundle(self, base_rev_id, rev_id, checkout_dir=None):
        """Create a bundle from base_rev_id -> rev_id in built-in branch.
        Make sure that the text generated is valid, and that it
        can be applied against the base, and generate the same information.

        :return: The in-memory bundle
        """
        bundle_txt, rev_ids = self.create_bundle_text(base_rev_id, rev_id)

        # This should also validate the generated bundle
        bundle = read_bundle(bundle_txt)
        repository = self.b1.repository
        for bundle_rev in bundle.real_revisions:
            # These really should have already been checked when we read the
            # bundle, since it computes the sha1 hash for the revision, which
            # only will match if everything is okay, but lets be explicit about
            # it
            branch_rev = repository.get_revision(bundle_rev.revision_id)
            for a in ('inventory_sha1', 'revision_id', 'parent_ids',
                      'timestamp', 'timezone', 'message', 'committer',
                      'parent_ids', 'properties'):
                self.assertEqual(getattr(branch_rev, a),
                                 getattr(bundle_rev, a))
            self.assertEqual(len(branch_rev.parent_ids),
                             len(bundle_rev.parent_ids))
        self.assertEqual(set(rev_ids),
                         {r.revision_id for r in bundle.real_revisions})
        self.valid_apply_bundle(base_rev_id, bundle,
                                checkout_dir=checkout_dir)

        return bundle

    def get_invalid_bundle(self, base_rev_id, rev_id):
        """Create a bundle from base_rev_id -> rev_id in built-in branch.
        Munge the text so that it's invalid.

        :return: The in-memory bundle
        """
        from ..bundle import serializer
        bundle_txt, rev_ids = self.create_bundle_text(base_rev_id, rev_id)
        new_text = self.get_raw(BytesIO(b''.join(bundle_txt)))
        new_text = new_text.replace(b'<file file_id="exe-1"',
                                    b'<file executable="y" file_id="exe-1"')
        new_text = new_text.replace(b'B260', b'B275')
        bundle_txt = BytesIO()
        bundle_txt.write(serializer._get_bundle_header('4'))
        bundle_txt.write(b'\n')
        bundle_txt.write(bz2.compress(new_text))
        bundle_txt.seek(0)
        bundle = read_bundle(bundle_txt)
        self.valid_apply_bundle(base_rev_id, bundle)
        return bundle

    def create_bundle_text(self, base_rev_id, rev_id):
        bundle_txt = BytesIO()
        rev_ids = write_bundle(self.b1.repository, rev_id, base_rev_id,
                               bundle_txt, format=self.format)
        bundle_txt.seek(0)
        self.assertEqual(bundle_txt.readline(),
                         b'# Bazaar revision bundle v%s\n' % self.format.encode('ascii'))
        self.assertEqual(bundle_txt.readline(), b'#\n')
        rev = self.b1.repository.get_revision(rev_id)
        bundle_txt.seek(0)
        return bundle_txt, rev_ids

    def get_bundle_tree(self, bundle, revision_id):
        repository = self.make_repository('repo')
        bundle.install_revisions(repository)
        return repository.revision_tree(revision_id)

    def test_creation(self):
        tree = self.make_branch_and_tree('tree')
        self.build_tree_contents([('tree/file', b'contents1\nstatic\n')])
        tree.add('file', ids=b'fileid-2')
        tree.commit('added file', rev_id=b'rev1')
        self.build_tree_contents([('tree/file', b'contents2\nstatic\n')])
        tree.commit('changed file', rev_id=b'rev2')
        s = BytesIO()
        serializer = BundleSerializerV4('1.0')
        with tree.lock_read():
            serializer.write_bundle(
                tree.branch.repository, b'rev2', b'null:', s)
        s.seek(0)
        tree2 = self.make_branch_and_tree('target')
        target_repo = tree2.branch.repository
        install_bundle(target_repo, serializer.read(s))
        target_repo.lock_read()
        self.addCleanup(target_repo.unlock)
        # Turn the 'iterators_of_bytes' back into simple strings for comparison
        repo_texts = {i: b''.join(content) for i, content
                          in target_repo.iter_files_bytes(
            [(b'fileid-2', b'rev1', '1'),
             (b'fileid-2', b'rev2', '2')])}
        self.assertEqual({'1': b'contents1\nstatic\n',
                          '2': b'contents2\nstatic\n'},
                         repo_texts)
        rtree = target_repo.revision_tree(b'rev2')
        inventory_vf = target_repo.inventories
        # If the inventory store has a graph, it must match the revision graph.
        self.assertSubset(
            [inventory_vf.get_parent_map([(b'rev2',)])[(b'rev2',)]],
            [None, ((b'rev1',),)])
        self.assertEqual('changed file',
                         target_repo.get_revision(b'rev2').message)

    @staticmethod
    def get_raw(bundle_file):
        bundle_file.seek(0)
        line = bundle_file.readline()
        line = bundle_file.readline()
        lines = bundle_file.readlines()
        return bz2.decompress(b''.join(lines))

    def test_copy_signatures(self):
        tree_a = self.make_branch_and_tree('tree_a')
        import breezy.commit as commit
        import breezy.gpg
        oldstrategy = breezy.gpg.GPGStrategy
        branch = tree_a.branch
        repo_a = branch.repository
        tree_a.commit("base", allow_pointless=True, rev_id=b'A')
        self.assertFalse(branch.repository.has_signature_for_revision_id(b'A'))
        try:
            from ..testament import Testament

            # monkey patch gpg signing mechanism
            breezy.gpg.GPGStrategy = breezy.gpg.LoopbackGPGStrategy
            new_config = test_commit.MustSignConfig()
            commit.Commit(config_stack=new_config).commit(message="base",
                                                          allow_pointless=True,
                                                          rev_id=b'B',
                                                          working_tree=tree_a)

            def sign(text):
                return breezy.gpg.LoopbackGPGStrategy(None).sign(text)
            self.assertTrue(repo_a.has_signature_for_revision_id(b'B'))
        finally:
            breezy.gpg.GPGStrategy = oldstrategy
        tree_b = self.make_branch_and_tree('tree_b')
        repo_b = tree_b.branch.repository
        s = BytesIO()
        serializer = BundleSerializerV4('4')
        with tree_a.lock_read():
            serializer.write_bundle(
                tree_a.branch.repository, b'B', b'null:', s)
        s.seek(0)
        install_bundle(repo_b, serializer.read(s))
        self.assertTrue(repo_b.has_signature_for_revision_id(b'B'))
        self.assertEqual(repo_b.get_signature_text(b'B'),
                         repo_a.get_signature_text(b'B'))
        s.seek(0)
        # ensure repeat installs are harmless
        install_bundle(repo_b, serializer.read(s))


class V4_2aBundleTester(V4BundleTester):

    def bzrdir_format(self):
        return '2a'

    def get_invalid_bundle(self, base_rev_id, rev_id):
        """Create a bundle from base_rev_id -> rev_id in built-in branch.
        Munge the text so that it's invalid.

        :return: The in-memory bundle
        """
        from ..bundle import serializer
        bundle_txt, rev_ids = self.create_bundle_text(base_rev_id, rev_id)
        new_text = self.get_raw(BytesIO(b''.join(bundle_txt)))
        # We are going to be replacing some text to set the executable bit on a
        # file. Make sure the text replacement actually works correctly.
        self.assertContainsRe(new_text, b'(?m)B244\n\ni 1\n<inventory')
        new_text = new_text.replace(b'<file file_id="exe-1"',
                                    b'<file executable="y" file_id="exe-1"')
        new_text = new_text.replace(b'B244', b'B259')
        bundle_txt = BytesIO()
        bundle_txt.write(serializer._get_bundle_header('4'))
        bundle_txt.write(b'\n')
        bundle_txt.write(bz2.compress(new_text))
        bundle_txt.seek(0)
        bundle = read_bundle(bundle_txt)
        self.valid_apply_bundle(base_rev_id, bundle)
        return bundle

    def make_merged_branch(self):
        builder = self.make_branch_builder('source')
        builder.start_series()
        builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', None)),
            ('add', ('file', b'file-id', 'file', b'original content\n')),
            ], revision_id=b'a@cset-0-1')
        builder.build_snapshot([b'a@cset-0-1'], [
            ('modify', ('file', b'new-content\n')),
            ], revision_id=b'a@cset-0-2a')
        builder.build_snapshot([b'a@cset-0-1'], [
            ('add', ('other-file', b'file2-id', 'file', b'file2-content\n')),
            ], revision_id=b'a@cset-0-2b')
        builder.build_snapshot([b'a@cset-0-2a', b'a@cset-0-2b'], [
            ('add', ('other-file', b'file2-id', 'file', b'file2-content\n')),
            ], revision_id=b'a@cset-0-3')
        builder.finish_series()
        self.b1 = builder.get_branch()
        self.b1.lock_read()
        self.addCleanup(self.b1.unlock)

    def make_bundle_just_inventories(self, base_revision_id,
                                     target_revision_id,
                                     revision_ids):
        sio = BytesIO()
        writer = v4.BundleWriteOperation(base_revision_id, target_revision_id,
                                         self.b1.repository, sio)
        writer.bundle.begin()
        writer._add_inventory_mpdiffs_from_serializer(revision_ids)
        writer.bundle.end()
        sio.seek(0)
        return sio

    def test_single_inventory_multiple_parents_as_xml(self):
        self.make_merged_branch()
        sio = self.make_bundle_just_inventories(b'a@cset-0-1', b'a@cset-0-3',
                                                [b'a@cset-0-3'])
        reader = v4.BundleReader(sio, stream_input=False)
        records = list(reader.iter_records())
        self.assertEqual(1, len(records))
        (bytes, metadata, repo_kind, revision_id,
         file_id) = records[0]
        self.assertIs(None, file_id)
        self.assertEqual(b'a@cset-0-3', revision_id)
        self.assertEqual('inventory', repo_kind)
        self.assertEqual({b'parents': [b'a@cset-0-2a', b'a@cset-0-2b'],
                          b'sha1': b'09c53b0c4de0895e11a2aacc34fef60a6e70865c',
                          b'storage_kind': b'mpdiff',
                          }, metadata)
        # We should have an mpdiff that takes some lines from both parents.
        self.assertEqualDiff(
            b'i 1\n'
            b'<inventory format="10" revision_id="a@cset-0-3">\n'
            b'\n'
            b'c 0 1 1 2\n'
            b'c 1 3 3 2\n', bytes)

    def test_single_inv_no_parents_as_xml(self):
        self.make_merged_branch()
        sio = self.make_bundle_just_inventories(b'null:', b'a@cset-0-1',
                                                [b'a@cset-0-1'])
        reader = v4.BundleReader(sio, stream_input=False)
        records = list(reader.iter_records())
        self.assertEqual(1, len(records))
        (bytes, metadata, repo_kind, revision_id,
         file_id) = records[0]
        self.assertIs(None, file_id)
        self.assertEqual(b'a@cset-0-1', revision_id)
        self.assertEqual('inventory', repo_kind)
        self.assertEqual({b'parents': [],
                          b'sha1': b'a13f42b142d544aac9b085c42595d304150e31a2',
                          b'storage_kind': b'mpdiff',
                          }, metadata)
        # We should have an mpdiff that takes some lines from both parents.
        self.assertEqualDiff(
            b'i 4\n'
            b'<inventory format="10" revision_id="a@cset-0-1">\n'
            b'<directory file_id="root-id" name=""'
            b' revision="a@cset-0-1" />\n'
            b'<file file_id="file-id" name="file" parent_id="root-id"'
            b' revision="a@cset-0-1"'
            b' text_sha1="09c2f8647e14e49e922b955c194102070597c2d1"'
            b' text_size="17" />\n'
            b'</inventory>\n'
            b'\n', bytes)

    def test_multiple_inventories_as_xml(self):
        self.make_merged_branch()
        sio = self.make_bundle_just_inventories(b'a@cset-0-1', b'a@cset-0-3',
                                                [b'a@cset-0-2a', b'a@cset-0-2b', b'a@cset-0-3'])
        reader = v4.BundleReader(sio, stream_input=False)
        records = list(reader.iter_records())
        self.assertEqual(3, len(records))
        revision_ids = [rev_id for b, m, k, rev_id, f in records]
        self.assertEqual([b'a@cset-0-2a', b'a@cset-0-2b', b'a@cset-0-3'],
                         revision_ids)
        metadata_2a = records[0][1]
        self.assertEqual({b'parents': [b'a@cset-0-1'],
                          b'sha1': b'1e105886d62d510763e22885eec733b66f5f09bf',
                          b'storage_kind': b'mpdiff',
                          }, metadata_2a)
        metadata_2b = records[1][1]
        self.assertEqual({b'parents': [b'a@cset-0-1'],
                          b'sha1': b'f03f12574bdb5ed2204c28636c98a8547544ccd8',
                          b'storage_kind': b'mpdiff',
                          }, metadata_2b)
        metadata_3 = records[2][1]
        self.assertEqual({b'parents': [b'a@cset-0-2a', b'a@cset-0-2b'],
                          b'sha1': b'09c53b0c4de0895e11a2aacc34fef60a6e70865c',
                          b'storage_kind': b'mpdiff',
                          }, metadata_3)
        bytes_2a = records[0][0]
        self.assertEqualDiff(
            b'i 1\n'
            b'<inventory format="10" revision_id="a@cset-0-2a">\n'
            b'\n'
            b'c 0 1 1 1\n'
            b'i 1\n'
            b'<file file_id="file-id" name="file" parent_id="root-id"'
            b' revision="a@cset-0-2a"'
            b' text_sha1="50f545ff40e57b6924b1f3174b267ffc4576e9a9"'
            b' text_size="12" />\n'
            b'\n'
            b'c 0 3 3 1\n', bytes_2a)
        bytes_2b = records[1][0]
        self.assertEqualDiff(
            b'i 1\n'
            b'<inventory format="10" revision_id="a@cset-0-2b">\n'
            b'\n'
            b'c 0 1 1 2\n'
            b'i 1\n'
            b'<file file_id="file2-id" name="other-file" parent_id="root-id"'
            b' revision="a@cset-0-2b"'
            b' text_sha1="b46c0c8ea1e5ef8e46fc8894bfd4752a88ec939e"'
            b' text_size="14" />\n'
            b'\n'
            b'c 0 3 4 1\n', bytes_2b)
        bytes_3 = records[2][0]
        self.assertEqualDiff(
            b'i 1\n'
            b'<inventory format="10" revision_id="a@cset-0-3">\n'
            b'\n'
            b'c 0 1 1 2\n'
            b'c 1 3 3 2\n', bytes_3)

    def test_creating_bundle_preserves_chk_pages(self):
        self.make_merged_branch()
        target = self.b1.controldir.sprout('target',
                                           revision_id=b'a@cset-0-2a').open_branch()
        bundle_txt, rev_ids = self.create_bundle_text(b'a@cset-0-2a',
                                                      b'a@cset-0-3')
        self.assertEqual({b'a@cset-0-2b', b'a@cset-0-3'}, set(rev_ids))
        bundle = read_bundle(bundle_txt)
        target.lock_write()
        self.addCleanup(target.unlock)
        install_bundle(target.repository, bundle)
        inv1 = next(self.b1.repository.inventories.get_record_stream([
            (b'a@cset-0-3',)], 'unordered',
            True)).get_bytes_as('fulltext')
        inv2 = next(target.repository.inventories.get_record_stream([
            (b'a@cset-0-3',)], 'unordered',
            True)).get_bytes_as('fulltext')
        self.assertEqualDiff(inv1, inv2)


class MungedBundleTester:

    def build_test_bundle(self):
        wt = self.make_branch_and_tree('b1')

        self.build_tree(['b1/one'])
        wt.add('one')
        wt.commit('add one', rev_id=b'a@cset-0-1')
        self.build_tree(['b1/two'])
        wt.add('two')
        wt.commit('add two', rev_id=b'a@cset-0-2',
                  revprops={'branch-nick': 'test'})

        bundle_txt = BytesIO()
        rev_ids = write_bundle(wt.branch.repository, b'a@cset-0-2',
                               b'a@cset-0-1', bundle_txt, self.format)
        self.assertEqual({b'a@cset-0-2'}, set(rev_ids))
        bundle_txt.seek(0, 0)
        return bundle_txt

    def check_valid(self, bundle):
        """Check that after whatever munging, the final object is valid."""
        self.assertEqual([b'a@cset-0-2'],
                         [r.revision_id for r in bundle.real_revisions])

    def test_extra_whitespace(self):
        bundle_txt = self.build_test_bundle()

        # Seek to the end of the file
        # Adding one extra newline used to give us
        # TypeError: float() argument must be a string or a number
        bundle_txt.seek(0, 2)
        bundle_txt.write(b'\n')
        bundle_txt.seek(0)

        bundle = read_bundle(bundle_txt)
        self.check_valid(bundle)

    def test_extra_whitespace_2(self):
        bundle_txt = self.build_test_bundle()

        # Seek to the end of the file
        # Adding two extra newlines used to give us
        # MalformedPatches: The first line of all patches should be ...
        bundle_txt.seek(0, 2)
        bundle_txt.write(b'\n\n')
        bundle_txt.seek(0)

        bundle = read_bundle(bundle_txt)
        self.check_valid(bundle)


class MungedBundleTesterV09(tests.TestCaseWithTransport, MungedBundleTester):

    format = '0.9'

    def test_missing_trailing_whitespace(self):
        bundle_txt = self.build_test_bundle()

        # Remove a trailing newline, it shouldn't kill the parser
        raw = bundle_txt.getvalue()
        # The contents of the bundle don't have to be this, but this
        # test is concerned with the exact case where the serializer
        # creates a blank line at the end, and fails if that
        # line is stripped
        self.assertEqual(b'\n\n', raw[-2:])
        bundle_txt = BytesIO(raw[:-1])

        bundle = read_bundle(bundle_txt)
        self.check_valid(bundle)

    def test_opening_text(self):
        bundle_txt = self.build_test_bundle()

        bundle_txt = BytesIO(
            b"Some random\nemail comments\n" + bundle_txt.getvalue())

        bundle = read_bundle(bundle_txt)
        self.check_valid(bundle)

    def test_trailing_text(self):
        bundle_txt = self.build_test_bundle()

        bundle_txt = BytesIO(
            bundle_txt.getvalue() + b"Some trailing\nrandom\ntext\n")

        bundle = read_bundle(bundle_txt)
        self.check_valid(bundle)


class MungedBundleTesterV4(tests.TestCaseWithTransport, MungedBundleTester):

    format = '4'


class TestBundleWriterReader(tests.TestCase):

    def test_roundtrip_record(self):
        fileobj = BytesIO()
        writer = v4.BundleWriter(fileobj)
        writer.begin()
        writer.add_info_record({b'foo': b'bar'})
        writer._add_record(b"Record body", {b'parents': [b'1', b'3'],
                                            b'storage_kind': b'fulltext'}, 'file', b'revid', b'fileid')
        writer.end()
        fileobj.seek(0)
        reader = v4.BundleReader(fileobj, stream_input=True)
        record_iter = reader.iter_records()
        record = next(record_iter)
        self.assertEqual((None, {b'foo': b'bar', b'storage_kind': b'header'},
                          'info', None, None), record)
        record = next(record_iter)
        self.assertEqual((b"Record body", {b'storage_kind': b'fulltext',
                                           b'parents': [b'1', b'3']}, 'file', b'revid', b'fileid'),
                         record)

    def test_roundtrip_record_memory_hungry(self):
        fileobj = BytesIO()
        writer = v4.BundleWriter(fileobj)
        writer.begin()
        writer.add_info_record({b'foo': b'bar'})
        writer._add_record(b"Record body", {b'parents': [b'1', b'3'],
                                            b'storage_kind': b'fulltext'}, 'file', b'revid', b'fileid')
        writer.end()
        fileobj.seek(0)
        reader = v4.BundleReader(fileobj, stream_input=False)
        record_iter = reader.iter_records()
        record = next(record_iter)
        self.assertEqual((None, {b'foo': b'bar', b'storage_kind': b'header'},
                          'info', None, None), record)
        record = next(record_iter)
        self.assertEqual((b"Record body", {b'storage_kind': b'fulltext',
                                           b'parents': [b'1', b'3']}, 'file', b'revid', b'fileid'),
                         record)

    def test_encode_name(self):
        self.assertEqual(b'revision/rev1',
                         v4.BundleWriter.encode_name('revision', b'rev1'))
        self.assertEqual(b'file/rev//1/file-id-1',
                         v4.BundleWriter.encode_name('file', b'rev/1', b'file-id-1'))
        self.assertEqual(b'info',
                         v4.BundleWriter.encode_name('info', None, None))

    def test_decode_name(self):
        self.assertEqual(('revision', b'rev1', None),
                         v4.BundleReader.decode_name(b'revision/rev1'))
        self.assertEqual(('file', b'rev/1', b'file-id-1'),
                         v4.BundleReader.decode_name(b'file/rev//1/file-id-1'))
        self.assertEqual(('info', None, None),
                         v4.BundleReader.decode_name(b'info'))

    def test_too_many_names(self):
        fileobj = BytesIO()
        writer = v4.BundleWriter(fileobj)
        writer.begin()
        writer.add_info_record({b'foo': b'bar'})
        writer._container.add_bytes_record([b'blah'], len(b'blah'), [(b'two', ), (b'names', )])
        writer.end()
        fileobj.seek(0)
        record_iter = v4.BundleReader(fileobj).iter_records()
        record = next(record_iter)
        self.assertEqual((None, {b'foo': b'bar', b'storage_kind': b'header'},
                          'info', None, None), record)
        self.assertRaises(errors.BadBundle, next, record_iter)



# Copyright (C) 2007 Canonical Ltd
# Authors:  Robert Collins <robert.collins@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from breezy.bzr import inventory, inventorytree
from breezy.tests import TestNotApplicable
from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestNestedSupport(TestCaseWithWorkingTree):

    def make_branch_and_tree(self, path):
        tree = TestCaseWithWorkingTree.make_branch_and_tree(self, path)
        if not tree.supports_tree_reference():
            raise TestNotApplicable('Tree references not supported')
        return tree

    def test_set_get_inventory_tree_reference(self):
        """This tests that setting a tree reference is persistent."""
        tree = self.make_branch_and_tree('.')
        if not isinstance(tree, inventorytree.InventoryTree):
            raise TestNotApplicable('not an inventory tree')
        transform = tree.transform()
        trans_id = transform.new_directory('reference', transform.root,
                                           b'subtree-id')
        transform.set_tree_reference(b'subtree-revision', trans_id)
        transform.apply()
        tree = tree.controldir.open_workingtree()
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual(b'subtree-revision',
                         tree.root_inventory.get_entry(b'subtree-id').reference_revision)

    def test_extract_while_locked(self):
        tree = self.make_branch_and_tree('.')
        tree.lock_write()
        self.addCleanup(tree.unlock)
        self.build_tree(['subtree/'])
        tree.add(['subtree'])
        subtree = tree.extract('subtree')

    def prepare_with_subtree(self):
        tree = self.make_branch_and_tree('.')
        tree.lock_write()
        self.addCleanup(tree.unlock)
        subtree = self.make_branch_and_tree('subtree')
        subtree.commit('dummy')
        tree.add(['subtree'])
        return tree

    def test_comparison_data_does_not_autodetect_subtree(self):
        tree = self.prepare_with_subtree()
        (path, versioned, kind, ie) = list(tree.list_files('subtree'))[0]
        self.assertEqual('directory',
                         tree._comparison_data(ie, 'subtree')[0])

    def test_may_not_autodetect_subtree(self):
        tree = self.prepare_with_subtree()
        self.assertIn(tree.kind('subtree'), ('directory', 'tree-reference'))

    def test_iter_entries_by_dir_autodetects_subtree(self):
        tree = self.prepare_with_subtree()
        path, ie = next(tree.iter_entries_by_dir(specific_files=['subtree']))
        self.assertEqual('tree-reference', ie.kind)



# Copyright (C) 2008-2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for finding, parsing and searching rule-based preferences."""

import sys

from breezy import rules, tests


class TestErrors(tests.TestCase):

    def test_unknown_rules(self):
        err = rules.UnknownRules(['foo', 'bar'])
        self.assertEqual("Unknown rules detected: foo, bar.", str(err))


class TestIniBasedRulesSearcher(tests.TestCase):

    def make_searcher(self, text):
        """Make a _RulesSearcher from a string"""
        if text is None:
            lines = None
        else:
            lines = text.splitlines()
        return rules._IniBasedRulesSearcher(lines)

    def test_unknown_namespace(self):
        self.assertRaises(rules.UnknownRules, rules._IniBasedRulesSearcher,
                          ["[junk]", "foo=bar"])

    def test_get_items_file_missing(self):
        rs = self.make_searcher(None)
        self.assertEqual((), rs.get_items('a.txt'))
        self.assertEqual((), rs.get_selected_items('a.txt', ['foo']))
        self.assertEqual(None, rs.get_single_value('a.txt', 'foo'))

    def test_get_items_file_empty(self):
        rs = self.make_searcher("")
        self.assertEqual((), rs.get_items('a.txt'))
        self.assertEqual((), rs.get_selected_items('a.txt', ['foo']))
        self.assertEqual(None, rs.get_single_value('a.txt', 'foo'))

    def test_get_items_from_extension_match(self):
        rs = self.make_searcher("[name *.txt]\nfoo=bar\na=True\n")
        self.assertEqual((), rs.get_items('a.py'))
        self.assertEqual((('foo', 'bar'), ('a', 'True')),
                         rs.get_items('a.txt'))
        self.assertEqual((('foo', 'bar'), ('a', 'True')),
                         rs.get_items('dir/a.txt'))
        self.assertEqual((('foo', 'bar'),),
                         rs.get_selected_items('a.txt', ['foo']))
        self.assertEqual('bar', rs.get_single_value('a.txt', 'foo'))

    def test_get_items_from_multiple_glob_match(self):
        rs = self.make_searcher(
            "[name *.txt *.py 'x x' \"y y\"]\nfoo=bar\na=True\n")
        self.assertEqual((), rs.get_items('NEWS'))
        self.assertEqual((('foo', 'bar'), ('a', 'True')),
                         rs.get_items('a.py'))
        self.assertEqual((('foo', 'bar'), ('a', 'True')),
                         rs.get_items('a.txt'))
        self.assertEqual((('foo', 'bar'), ('a', 'True')),
                         rs.get_items('x x'))
        self.assertEqual((('foo', 'bar'), ('a', 'True')),
                         rs.get_items('y y'))
        self.assertEqual('bar', rs.get_single_value('a.txt', 'foo'))

    def test_get_items_pathname_match(self):
        rs = self.make_searcher("[name ./a.txt]\nfoo=baz\n")
        self.assertEqual((('foo', 'baz'),),
                         rs.get_items('a.txt'))
        self.assertEqual('baz', rs.get_single_value('a.txt', 'foo'))
        self.assertEqual((), rs.get_items('dir/a.txt'))
        self.assertEqual(None, rs.get_single_value('dir/a.txt', 'foo'))

    def test_get_items_match_first(self):
        rs = self.make_searcher(
            "[name ./a.txt]\nfoo=baz\n"
            "[name *.txt]\nfoo=bar\na=True\n")
        self.assertEqual((('foo', 'baz'),),
                         rs.get_items('a.txt'))
        self.assertEqual('baz', rs.get_single_value('a.txt', 'foo'))
        self.assertEqual((('foo', 'bar'), ('a', 'True')),
                         rs.get_items('dir/a.txt'))
        self.assertEqual('bar', rs.get_single_value('dir/a.txt', 'foo'))


class TestStackedRulesSearcher(tests.TestCase):

    def make_searcher(self, text1=None, text2=None):
        """Make a _StackedRulesSearcher with 0, 1 or 2 items"""
        searchers = []
        if text1 is not None:
            searchers.append(rules._IniBasedRulesSearcher(
                text1.splitlines()))
        if text2 is not None:
            searchers.append(rules._IniBasedRulesSearcher(
                text2.splitlines()))
        return rules._StackedRulesSearcher(searchers)

    def test_stack_searching(self):
        rs = self.make_searcher(
            "[name ./a.txt]\nfoo=baz\n",
            "[name *.txt]\nfoo=bar\na=True\n")
        self.assertEqual((('foo', 'baz'),),
                         rs.get_items('a.txt'))
        self.assertEqual('baz', rs.get_single_value('a.txt', 'foo'))
        self.assertEqual(None, rs.get_single_value('a.txt', 'a'))
        self.assertEqual((('foo', 'bar'), ('a', 'True')),
                         rs.get_items('dir/a.txt'))
        self.assertEqual('bar', rs.get_single_value('dir/a.txt', 'foo'))
        self.assertEqual('True', rs.get_single_value('dir/a.txt', 'a'))


class TestRulesPath(tests.TestCase):

    def setUp(self):
        super().setUp()
        self.overrideEnv('HOME', '/home/bogus')
        if sys.platform == 'win32':
            self.overrideEnv(
                'BRZ_HOME', r'C:\Documents and Settings\bogus\Application Data')
            self.brz_home = \
                'C:/Documents and Settings/bogus/Application Data/breezy'
        else:
            self.brz_home = '/home/bogus/.config/breezy'

    def test_rules_path(self):
        self.assertEqual(rules.rules_path(), self.brz_home + '/rules')



# Copyright (C) 2006, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test that brz handles locales in a reasonable way"""

import sys
import unittest

from breezy import tests


class TestLocale(tests.TestCaseWithTransport):

    def setUp(self):
        super().setUp()

        if sys.platform in ('win32',):
            raise tests.TestSkipped('Windows does not respond to the LANG'
                                    ' env variable')

        tree = self.make_branch_and_tree('tree')
        self.build_tree(['tree/a'])
        tree.add('a')
        tree.commit('Unicode \xb5 commit', rev_id=b'r1',
                    committer='\u062c\u0648\u062c\u0648'
                              ' Meinel <juju@info.com>',
                    timestamp=1156451297.96, timezone=0)
        self.tree = tree

    def run_log_quiet_long(self, args, env_changes={}):
        cmd = ['--no-aliases', '--no-plugins', '-Oprogress_bar=none',
               'log', '-q', '--log-format=long']
        cmd.extend(args)
        return self.run_brz_subprocess(cmd, env_changes=env_changes)

    @unittest.skip("encoding when LANG=C is currently borked")
    def test_log_coerced_utf8(self):
        self.disable_missing_extensions_warning()
        out, err = self.run_log_quiet_long(
            ['tree'],
            # C is not necessarily the default locale, so set both LANG and
            # LC_ALL explicitly because LC_ALL is preferred on (some?) Linux
            # systems but only LANG is respected on Windows.
            env_changes={
                'LANG': 'C', 'LC_ALL': 'C', 'LC_CTYPE': None, 'LANGUAGE': None})
        self.assertEqual(b'', err)
        self.assertEqualDiff(b"""\
------------------------------------------------------------
revno: 1
committer: \xd8\xac\xd9\x88\xd8\xac\xd9\x88 Meinel <juju@info.com>
branch nick: tree
timestamp: Thu 2006-08-24 20:28:17 +0000
message:
  Unicode \xc2\xb5 commit
""", out)

    @unittest.skipIf(
        sys.version_info[:2] >= (3, 8),
        "python > 3.8 doesn't allow changing filesystem default encoding")
    def test_log_C(self):
        self.disable_missing_extensions_warning()
        out, err = self.run_log_quiet_long(
            ['tree'],
            # C is not necessarily the default locale, so set both LANG and
            # LC_ALL explicitly because LC_ALL is preferred on (some?) Linux
            # systems but only LANG is respected on Windows.
            env_changes={
                'LANG': 'C', 'LC_ALL': 'C', 'LC_CTYPE': None,
                'LANGUAGE': None, 'PYTHONCOERCECLOCALE': '0', 'PYTHONUTF8': '0'})
        self.assertEqual(b'', err)
        self.assertEqualDiff(b"""\
------------------------------------------------------------
revno: 1
committer: ???? Meinel <juju@info.com>
branch nick: tree
timestamp: Thu 2006-08-24 20:28:17 +0000
message:
  Unicode ? commit
""", out)

    @unittest.skipIf(
        sys.version_info[:2] >= (3, 8),
        "python > 3.8 doesn't allow changing filesystem default encoding")
    def test_log_BOGUS(self):
        out, err = self.run_log_quiet_long(
            ['tree'],
            env_changes={'LANG': 'BOGUS', 'LC_ALL': None, 'LC_CTYPE': None,
                         'LANGUAGE': None, 'PYTHONCOERCECLOCALE': '0',
                         'PYTHONUTF8': '0'})
        self.assertStartsWith(err, b'brz: WARNING: Error: unsupported locale setting')
        self.assertEqualDiff(b"""\
------------------------------------------------------------
revno: 1
committer: ???? Meinel <juju@info.com>
branch nick: tree
timestamp: Thu 2006-08-24 20:28:17 +0000
message:
  Unicode ? commit
""", out)


class TestMultibyteCodecs(tests.TestCaseWithTransport):
    """Tests for quirks of multibyte encodings and their python codecs"""

    def test_plugins_mbcs(self):
        """Ensure the plugins command works with cjkcodecs, see lp:754082"""
        self.disable_missing_extensions_warning()
        out, err = self.run_bzr(["plugins"], encoding="EUC-JP")
        # The output is tested in bt.test_plugins rather than here
        self.assertEqual("", err)



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for repositories that do not support CHK indices.

CHK support is optional, and when it is not supported the methods and
attributes CHK support added should fail in known ways.
"""

from breezy.bzr.tests.per_repository_chk import TestCaseWithRepositoryCHK


class TestNoCHKSupport(TestCaseWithRepositoryCHK):

    def test_chk_bytes_attribute_is_None(self):
        repo = self.make_repository('.')
        self.assertEqual(None, repo.chk_bytes)



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for Repository.add_inventory_by_delta."""

from breezy import errors, revision
from breezy.bzr.tests.per_repository_vf import (
    TestCaseWithRepository, all_repository_vf_format_scenarios)
from breezy.tests.scenarios import load_tests_apply_scenarios

load_tests = load_tests_apply_scenarios


class TestAddInventoryByDelta(TestCaseWithRepository):

    scenarios = all_repository_vf_format_scenarios()

    def _get_repo_in_write_group(self, path='repository'):
        repo = self.make_repository(path)
        repo.lock_write()
        self.addCleanup(repo.unlock)
        repo.start_write_group()
        return repo

    def test_basis_missing_errors(self):
        repo = self._get_repo_in_write_group()
        try:
            self.assertRaises(errors.NoSuchRevision,
                              repo.add_inventory_by_delta, "missing-revision", [],
                              "new-revision", ["missing-revision"])
        finally:
            repo.abort_write_group()

    def test_not_in_write_group_errors(self):
        repo = self.make_repository('repository')
        repo.lock_write()
        self.addCleanup(repo.unlock)
        self.assertRaises(AssertionError, repo.add_inventory_by_delta,
                          "missing-revision", [], "new-revision", ["missing-revision"])

    def make_inv_delta(self, old, new):
        """Make an inventory delta from two inventories."""
        by_id = getattr(old, '_byid', None)
        if by_id is None:
            old_ids = {entry.file_id for entry in old.iter_just_entries()}
        else:
            old_ids = set(by_id)
        by_id = getattr(new, '_byid', None)
        if by_id is None:
            new_ids = {entry.file_id for entry in new.iter_just_entries()}
        else:
            new_ids = set(by_id)

        adds = new_ids - old_ids
        deletes = old_ids - new_ids
        common = old_ids.intersection(new_ids)
        delta = []
        for file_id in deletes:
            delta.append((old.id2path(file_id), None, file_id, None))
        for file_id in adds:
            delta.append((None, new.id2path(file_id),
                          file_id, new.get_entry(file_id)))
        for file_id in common:
            if old.get_entry(file_id) != new.get_entry(file_id):
                delta.append((old.id2path(file_id), new.id2path(file_id),
                              file_id, new[file_id]))
        return delta

    def test_same_validator(self):
        # Adding an inventory via delta or direct results in the same
        # validator.
        tree = self.make_branch_and_tree('tree')
        revid = tree.commit("empty post")
        # tree.basis_tree() always uses a plain Inventory from the dirstate, we
        # want the same format inventory as we have in the repository
        revtree = tree.branch.repository.revision_tree(
            tree.branch.last_revision())
        tree.basis_tree()
        revtree.lock_read()
        self.addCleanup(revtree.unlock)
        old_inv = tree.branch.repository.revision_tree(
            revision.NULL_REVISION).root_inventory
        new_inv = revtree.root_inventory
        delta = self.make_inv_delta(old_inv, new_inv)
        repo_direct = self._get_repo_in_write_group('direct')
        add_validator = repo_direct.add_inventory(revid, new_inv, [])
        repo_direct.commit_write_group()
        repo_delta = self._get_repo_in_write_group('delta')
        try:
            delta_validator, inv = repo_delta.add_inventory_by_delta(
                revision.NULL_REVISION, delta, revid, [])
        except:
            repo_delta.abort_write_group()
            raise
        else:
            repo_delta.commit_write_group()
        self.assertEqual(add_validator, delta_validator)
        self.assertEqual(list(new_inv.iter_entries()),
                         list(inv.iter_entries()))



# Copyright (C) 2006-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Black-box tests for brz log."""


import os

from breezy import branchbuilder, errors, log, osutils, tests
from breezy.tests import features, test_log


class TestLog(tests.TestCaseWithTransport, test_log.TestLogMixin):

    def make_minimal_branch(self, path='.', format=None):
        tree = self.make_branch_and_tree(path, format=format)
        self.build_tree([path + '/hello.txt'])
        tree.add('hello.txt')
        tree.commit(message='message1')
        return tree

    def make_linear_branch(self, path='.', format=None):
        tree = self.make_branch_and_tree(path, format=format)
        self.build_tree(
            [path + '/hello.txt', path + '/goodbye.txt', path + '/meep.txt'])
        tree.add('hello.txt')
        tree.commit(message='message1')
        tree.add('goodbye.txt')
        tree.commit(message='message2')
        tree.add('meep.txt')
        tree.commit(message='message3')
        return tree

    def make_merged_branch(self, path='.', format=None):
        tree = self.make_linear_branch(path, format)
        tree2 = tree.controldir.sprout('tree2',
                                       revision_id=tree.branch.get_rev_id(1)).open_workingtree()
        tree2.commit(message='tree2 message2')
        tree2.commit(message='tree2 message3')
        tree.merge_from_branch(tree2.branch)
        tree.commit(message='merge')
        return tree


class TestLogWithLogCatcher(TestLog):

    def setUp(self):
        super().setUp()
        # Capture log formatter creations

        class MyLogFormatter(test_log.LogCatcher):

            def __new__(klass, *args, **kwargs):
                self.log_catcher = test_log.LogCatcher(*args, **kwargs)
                # Always return our own log formatter
                return self.log_catcher
        # Break cycle with closure over self on cleanup by removing method
        self.addCleanup(setattr, MyLogFormatter, "__new__", None)

        def getme(branch):
            # Always return our own log formatter class hijacking the
            # default behavior (which requires setting up a config
            # variable)
            return MyLogFormatter
        self.overrideAttr(log.log_formatter_registry, 'get_default', getme)

    def get_captured_revisions(self):
        return self.log_catcher.revisions

    def assertLogRevnos(self, args, expected_revnos, working_dir='.',
                        out='', err=''):
        actual_out, actual_err = self.run_bzr(['log'] + args,
                                              working_dir=working_dir)
        self.assertEqual(out, actual_out)
        self.assertEqual(err, actual_err)
        self.assertEqual(expected_revnos,
                         [r.revno for r in self.get_captured_revisions()])

    def assertLogRevnosAndDepths(self, args, expected_revnos_and_depths,
                                 working_dir='.'):
        self.run_bzr(['log'] + args, working_dir=working_dir)
        self.assertEqual(expected_revnos_and_depths,
                         [(r.revno, r.merge_depth)
                          for r in self.get_captured_revisions()])


class TestLogRevSpecs(TestLogWithLogCatcher):

    def test_log_no_revspec(self):
        self.make_linear_branch()
        self.assertLogRevnos([], ['3', '2', '1'])

    def test_log_null_end_revspec(self):
        self.make_linear_branch()
        self.assertLogRevnos(['-r1..'], ['3', '2', '1'])

    def test_log_null_begin_revspec(self):
        self.make_linear_branch()
        self.assertLogRevnos(['-r..3'], ['3', '2', '1'])

    def test_log_null_both_revspecs(self):
        self.make_linear_branch()
        self.assertLogRevnos(['-r..'], ['3', '2', '1'])

    def test_log_negative_begin_revspec_full_log(self):
        self.make_linear_branch()
        self.assertLogRevnos(['-r-3..'], ['3', '2', '1'])

    def test_log_negative_both_revspec_full_log(self):
        self.make_linear_branch()
        self.assertLogRevnos(['-r-3..-1'], ['3', '2', '1'])

    def test_log_negative_both_revspec_partial(self):
        self.make_linear_branch()
        self.assertLogRevnos(['-r-3..-2'], ['2', '1'])

    def test_log_negative_begin_revspec(self):
        self.make_linear_branch()
        self.assertLogRevnos(['-r-2..'], ['3', '2'])

    def test_log_positive_revspecs(self):
        self.make_linear_branch()
        self.assertLogRevnos(['-r1..3'], ['3', '2', '1'])

    def test_log_dotted_revspecs(self):
        self.make_merged_branch()
        self.assertLogRevnos(['-n0', '-r1..1.1.1'], ['1.1.1', '1'])

    def test_log_limit(self):
        tree = self.make_branch_and_tree('.')
        # We want more commits than our batch size starts at
        for pos in range(10):
            tree.commit("%s" % pos)
        self.assertLogRevnos(['--limit', '2'], ['10', '9'])

    def test_log_limit_short(self):
        self.make_linear_branch()
        self.assertLogRevnos(['-l', '2'], ['3', '2'])

    def test_log_change_revno(self):
        self.make_linear_branch()
        self.assertLogRevnos(['-c1'], ['1'])

    def test_branch_revspec(self):
        foo = self.make_branch_and_tree('foo')
        bar = self.make_branch_and_tree('bar')
        self.build_tree(['foo/foo.txt', 'bar/bar.txt'])
        foo.add('foo.txt')
        bar.add('bar.txt')
        foo.commit(message='foo')
        bar.commit(message='bar')
        self.run_bzr('log -r branch:../bar', working_dir='foo')
        self.assertEqual([bar.branch.get_rev_id(1)],
                         [r.rev.revision_id
                          for r in self.get_captured_revisions()])


class TestLogExcludeCommonAncestry(TestLogWithLogCatcher):

    def test_exclude_common_ancestry_simple_revnos(self):
        self.make_linear_branch()
        self.assertLogRevnos(['-r1..3', '--exclude-common-ancestry'],
                             ['3', '2'])


class TestLogMergedLinearAncestry(TestLogWithLogCatcher):

    def setUp(self):
        super().setUp()
        # FIXME: Using a MemoryTree would be even better here (but until we
        # stop calling run_bzr, there is no point) --vila 100118.
        builder = branchbuilder.BranchBuilder(self.get_transport())
        builder.start_series()
        # 1
        # | \
        # 2  1.1.1
        # | / |
        # 3  1.1.2
        # |   |
        # |  1.1.3
        # | / |
        # 4  1.1.4
        # | /
        # 5
        # | \
        # | 5.1.1
        # | /
        # 6

        # mainline
        builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', ''))],
            revision_id=b'1')
        builder.build_snapshot([b'1'], [], revision_id=b'2')
        # branch
        builder.build_snapshot([b'1'], [], revision_id=b'1.1.1')
        # merge branch into mainline
        builder.build_snapshot([b'2', b'1.1.1'], [], revision_id=b'3')
        # new commits in branch
        builder.build_snapshot([b'1.1.1'], [], revision_id=b'1.1.2')
        builder.build_snapshot([b'1.1.2'], [], revision_id=b'1.1.3')
        # merge branch into mainline
        builder.build_snapshot([b'3', b'1.1.3'], [], revision_id=b'4')
        # merge mainline into branch
        builder.build_snapshot([b'1.1.3', b'4'], [], revision_id=b'1.1.4')
        # merge branch into mainline
        builder.build_snapshot([b'4', b'1.1.4'], [], revision_id=b'5')
        builder.build_snapshot([b'5'], [], revision_id=b'5.1.1')
        builder.build_snapshot([b'5', b'5.1.1'], [], revision_id=b'6')
        builder.finish_series()

    def test_n0(self):
        self.assertLogRevnos(['-n0', '-r1.1.1..1.1.4'],
                             ['1.1.4', '4', '1.1.3', '1.1.2', '3', '1.1.1'])

    def test_n0_forward(self):
        self.assertLogRevnos(['-n0', '-r1.1.1..1.1.4', '--forward'],
                             ['3', '1.1.1', '4', '1.1.2', '1.1.3', '1.1.4'])

    def test_n1(self):
        # starting from 1.1.4 we follow the left-hand ancestry
        self.assertLogRevnos(['-n1', '-r1.1.1..1.1.4'],
                             ['1.1.4', '1.1.3', '1.1.2', '1.1.1'])

    def test_n1_forward(self):
        self.assertLogRevnos(['-n1', '-r1.1.1..1.1.4', '--forward'],
                             ['1.1.1', '1.1.2', '1.1.3', '1.1.4'])

    def test_fallback_when_end_rev_is_not_on_mainline(self):
        self.assertLogRevnos(['-n1', '-r1.1.1..5.1.1'],
                             # We don't get 1.1.1 because we say -n1
                             ['5.1.1', '5', '4', '3'])


class Test_GenerateAllRevisions(TestLogWithLogCatcher):

    def setUp(self):
        super().setUp()
        builder = self.make_branch_with_many_merges()
        b = builder.get_branch()
        b.lock_read()
        self.addCleanup(b.unlock)
        self.branch = b

    def make_branch_with_many_merges(self, path='.', format=None):
        builder = branchbuilder.BranchBuilder(self.get_transport())
        builder.start_series()
        # The graph below may look a bit complicated (and it may be but I've
        # banged my head enough on it) but the bug requires at least dotted
        # revnos *and* merged revisions below that.
        # 1
        # | \
        # 2  1.1.1
        # | X
        # 3  2.1.1
        # |   |    \
        # |  2.1.2  2.2.1
        # |   |    X
        # |  2.1.3  \
        # | /       /
        # 4        /
        # |       /
        # 5 -----/
        builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', ''))], revision_id=b'1')
        builder.build_snapshot([b'1'], [], revision_id=b'2')
        builder.build_snapshot([b'1'], [], revision_id=b'1.1.1')
        builder.build_snapshot([b'2'], [], revision_id=b'2.1.1')
        builder.build_snapshot([b'2', b'1.1.1'], [], revision_id=b'3')
        builder.build_snapshot([b'2.1.1'], [], revision_id=b'2.1.2')
        builder.build_snapshot([b'2.1.1'], [], revision_id=b'2.2.1')
        builder.build_snapshot([b'2.1.2', b'2.2.1'], [], revision_id=b'2.1.3')
        builder.build_snapshot([b'3', b'2.1.3'], [], revision_id=b'4')
        builder.build_snapshot([b'4', b'2.1.2'], [], revision_id=b'5')
        builder.finish_series()
        return builder

    def test_not_an_ancestor(self):
        self.assertRaises(errors.CommandError,
                          log._generate_all_revisions,
                          self.branch, '1.1.1', '2.1.3', 'reverse',
                          delayed_graph_generation=True)

    def test_wrong_order(self):
        self.assertRaises(errors.CommandError,
                          log._generate_all_revisions,
                          self.branch, '5', '2.1.3', 'reverse',
                          delayed_graph_generation=True)

    def test_no_start_rev_id_with_end_rev_id_being_a_merge(self):
        revs = log._generate_all_revisions(
            self.branch, None, '2.1.3',
            'reverse', delayed_graph_generation=True)


class TestLogRevSpecsWithPaths(TestLogWithLogCatcher):

    def test_log_revno_n_path_wrong_namespace(self):
        self.make_linear_branch('branch1')
        self.make_linear_branch('branch2')
        # There is no guarantee that a path exist between two arbitrary
        # revisions.
        self.run_bzr("log -r revno:2:branch1..revno:3:branch2", retcode=3)

    def test_log_revno_n_path_correct_order(self):
        self.make_linear_branch('branch2')
        self.assertLogRevnos(['-rrevno:1:branch2..revno:3:branch2'],
                             ['3', '2', '1'])

    def test_log_revno_n_path(self):
        self.make_linear_branch('branch2')
        self.assertLogRevnos(['-rrevno:1:branch2'],
                             ['1'])
        rev_props = self.log_catcher.revisions[0].rev.properties
        self.assertEqual('branch2', rev_props['branch-nick'])


class TestLogErrors(TestLog):

    def test_log_zero_revspec(self):
        self.make_minimal_branch()
        self.run_bzr_error(['brz: ERROR: Logging revision 0 is invalid.'],
                           ['log', '-r0'])

    def test_log_zero_begin_revspec(self):
        self.make_linear_branch()
        self.run_bzr_error(['brz: ERROR: Logging revision 0 is invalid.'],
                           ['log', '-r0..2'])

    def test_log_zero_end_revspec(self):
        self.make_linear_branch()
        self.run_bzr_error(['brz: ERROR: Logging revision 0 is invalid.'],
                           ['log', '-r-2..0'])

    def test_log_nonexistent_revno(self):
        self.make_minimal_branch()
        self.run_bzr_error(["brz: ERROR: Requested revision: '1234' "
                            "does not exist in branch:"],
                           ['log', '-r1234'])

    def test_log_nonexistent_dotted_revno(self):
        self.make_minimal_branch()
        self.run_bzr_error(["brz: ERROR: Requested revision: '123.123' "
                            "does not exist in branch:"],
                           ['log', '-r123.123'])

    def test_log_change_nonexistent_revno(self):
        self.make_minimal_branch()
        self.run_bzr_error(["brz: ERROR: Requested revision: '1234' "
                            "does not exist in branch:"],
                           ['log', '-c1234'])

    def test_log_change_nonexistent_dotted_revno(self):
        self.make_minimal_branch()
        self.run_bzr_error(["brz: ERROR: Requested revision: '123.123' "
                            "does not exist in branch:"],
                           ['log', '-c123.123'])

    def test_log_change_single_revno_only(self):
        self.make_minimal_branch()
        self.run_bzr_error(['brz: ERROR: Option --change does not'
                            ' accept revision ranges'],
                           ['log', '--change', '2..3'])

    def test_log_change_incompatible_with_revision(self):
        self.run_bzr_error(['brz: ERROR: --revision and --change'
                            ' are mutually exclusive'],
                           ['log', '--change', '2', '--revision', '3'])

    def test_log_nonexistent_file(self):
        self.make_minimal_branch()
        # files that don't exist in either the basis tree or working tree
        # should give an error
        out, err = self.run_bzr('log does-not-exist', retcode=3)
        self.assertContainsRe(err,
                              'Path unknown at end or start of revision range: '
                              'does-not-exist')

    def test_log_reversed_revspecs(self):
        self.make_linear_branch()
        self.run_bzr_error(('brz: ERROR: Start revision must be older than '
                            'the end revision.\n',),
                           ['log', '-r3..1'])

    def test_log_reversed_dotted_revspecs(self):
        self.make_merged_branch()
        self.run_bzr_error(('brz: ERROR: Start revision not found in '
                            'history of end revision.\n',),
                           "log -r 1.1.1..1")

    def test_log_bad_message_re(self):
        """Bad --message argument gives a sensible message

        See https://bugs.launchpad.net/bzr/+bug/251352
        """
        self.make_minimal_branch()
        out, err = self.run_bzr(['log', '-m', '*'], retcode=3)
        self.assertContainsRe(err, "ERROR.*Invalid pattern.*nothing to repeat")
        self.assertNotContainsRe(err, "Unprintable exception")
        self.assertEqual(out, '')

    def test_log_unsupported_timezone(self):
        self.make_linear_branch()
        self.run_bzr_error(['brz: ERROR: Unsupported timezone format "foo", '
                            'options are "utc", "original", "local".'],
                           ['log', '--timezone', 'foo'])

    def test_log_exclude_ancestry_no_range(self):
        self.make_linear_branch()
        self.run_bzr_error(['brz: ERROR: --exclude-common-ancestry'
                            ' requires -r with two revisions'],
                           ['log', '--exclude-common-ancestry'])

    def test_log_exclude_ancestry_single_revision(self):
        self.make_merged_branch()
        self.run_bzr_error(['brz: ERROR: --exclude-common-ancestry'
                            ' requires two different revisions'],
                           ['log', '--exclude-common-ancestry',
                            '-r1.1.1..1.1.1'])


class TestLogTags(TestLog):

    def test_log_with_tags(self):
        tree = self.make_linear_branch(format='dirstate-tags')
        branch = tree.branch
        branch.tags.set_tag('tag1', branch.get_rev_id(1))
        branch.tags.set_tag('tag1.1', branch.get_rev_id(1))
        branch.tags.set_tag('tag3', branch.last_revision())

        log = self.run_bzr("log -r-1")[0]
        self.assertTrue('tags: tag3' in log)

        log = self.run_bzr("log -r1")[0]
        # I guess that we can't know the order of tags in the output
        # since dicts are unordered, need to check both possibilities
        self.assertContainsRe(log, r'tags: (tag1, tag1\.1|tag1\.1, tag1)')

    def test_merged_log_with_tags(self):
        branch1_tree = self.make_linear_branch('branch1',
                                               format='dirstate-tags')
        branch1 = branch1_tree.branch
        branch2_tree = branch1_tree.controldir.sprout(
            'branch2').open_workingtree()
        branch1_tree.commit(message='foobar', allow_pointless=True)
        branch1.tags.set_tag('tag1', branch1.last_revision())
        # tags don't propagate if we don't merge
        self.run_bzr('merge ../branch1', working_dir='branch2')
        branch2_tree.commit(message='merge branch 1')
        log = self.run_bzr("log -n0 -r-1", working_dir='branch2')[0]
        self.assertContainsRe(log, r'    tags: tag1')
        log = self.run_bzr("log -n0 -r3.1.1", working_dir='branch2')[0]
        self.assertContainsRe(log, r'tags: tag1')


class TestLogSignatures(TestLog):

    def test_log_with_signatures(self):
        self.requireFeature(features.gpg)

        tree = self.make_linear_branch(format='dirstate-tags')

        log = self.run_bzr("log --signatures")[0]
        self.assertTrue('signature: no signature' in log)

    def test_log_without_signatures(self):
        self.requireFeature(features.gpg)

        tree = self.make_linear_branch(format='dirstate-tags')

        log = self.run_bzr("log")[0]
        self.assertFalse('signature: no signature' in log)


class TestLogVerbose(TestLog):

    def setUp(self):
        super().setUp()
        self.make_minimal_branch()

    def assertUseShortDeltaFormat(self, cmd):
        log = self.run_bzr(cmd)[0]
        # Check that we use the short status format
        self.assertContainsRe(log, '(?m)^\\s*A  hello.txt$')
        self.assertNotContainsRe(log, '(?m)^\\s*added:$')

    def assertUseLongDeltaFormat(self, cmd):
        log = self.run_bzr(cmd)[0]
        # Check that we use the long status format
        self.assertNotContainsRe(log, '(?m)^\\s*A  hello.txt$')
        self.assertContainsRe(log, '(?m)^\\s*added:$')

    def test_log_short_verbose(self):
        self.assertUseShortDeltaFormat(['log', '--short', '-v'])

    def test_log_s_verbose(self):
        self.assertUseShortDeltaFormat(['log', '-S', '-v'])

    def test_log_short_verbose_verbose(self):
        self.assertUseLongDeltaFormat(['log', '--short', '-vv'])

    def test_log_long_verbose(self):
        # Check that we use the long status format, ignoring the verbosity
        # level
        self.assertUseLongDeltaFormat(['log', '--long', '-v'])

    def test_log_long_verbose_verbose(self):
        # Check that we use the long status format, ignoring the verbosity
        # level
        self.assertUseLongDeltaFormat(['log', '--long', '-vv'])


class TestLogMerges(TestLogWithLogCatcher):

    def setUp(self):
        super().setUp()
        self.make_branches_with_merges()

    def make_branches_with_merges(self):
        level0 = self.make_branch_and_tree('level0')
        self.wt_commit(level0, 'in branch level0')
        level1 = level0.controldir.sprout('level1').open_workingtree()
        self.wt_commit(level1, 'in branch level1')
        level2 = level1.controldir.sprout('level2').open_workingtree()
        self.wt_commit(level2, 'in branch level2')
        level1.merge_from_branch(level2.branch)
        self.wt_commit(level1, 'merge branch level2')
        level0.merge_from_branch(level1.branch)
        self.wt_commit(level0, 'merge branch level1')

    def test_merges_are_indented_by_level(self):
        self.run_bzr(['log', '-n0'], working_dir='level0')
        revnos_and_depth = [(r.revno, r.merge_depth)
                            for r in self.get_captured_revisions()]
        self.assertEqual([('2', 0), ('1.1.2', 1), ('1.2.1', 2), ('1.1.1', 1),
                          ('1', 0)],
                         [(r.revno, r.merge_depth)
                          for r in self.get_captured_revisions()])

    def test_force_merge_revisions_off(self):
        self.assertLogRevnos(['-n1'], ['2', '1'], working_dir='level0')

    def test_force_merge_revisions_on(self):
        self.assertLogRevnos(['-n0'], ['2', '1.1.2', '1.2.1', '1.1.1', '1'],
                             working_dir='level0')

    def test_include_merged(self):
        # Confirm --include-merged gives the same output as -n0
        expected = ['2', '1.1.2', '1.2.1', '1.1.1', '1']
        self.assertLogRevnos(['--include-merged'],
                             expected, working_dir='level0')
        self.assertLogRevnos(['--include-merged'],
                             expected, working_dir='level0')

    def test_force_merge_revisions_N(self):
        self.assertLogRevnos(['-n2'],
                             ['2', '1.1.2', '1.1.1', '1'],
                             working_dir='level0')

    def test_merges_single_merge_rev(self):
        self.assertLogRevnosAndDepths(['-n0', '-r1.1.2'],
                                      [('1.1.2', 0), ('1.2.1', 1)],
                                      working_dir='level0')

    def test_merges_partial_range(self):
        self.assertLogRevnosAndDepths(
            ['-n0', '-r1.1.1..1.1.2'],
            [('1.1.2', 0), ('1.2.1', 1), ('1.1.1', 0)],
            working_dir='level0')

    def test_merges_partial_range_ignore_before_lower_bound(self):
        """Dont show revisions before the lower bound's merged revs"""
        self.assertLogRevnosAndDepths(
            ['-n0', '-r1.1.2..2'],
            [('2', 0), ('1.1.2', 1), ('1.2.1', 2)],
            working_dir='level0')

    def test_omit_merges_with_sidelines(self):
        self.assertLogRevnos(['--omit-merges', '-n0'], ['1.2.1', '1.1.1', '1'],
                             working_dir='level0')

    def test_omit_merges_without_sidelines(self):
        self.assertLogRevnos(['--omit-merges', '-n1'], ['1'],
                             working_dir='level0')


class TestLogDiff(TestLogWithLogCatcher):

    # FIXME: We need specific tests for each LogFormatter about how the diffs
    # are displayed: --long indent them by depth, --short use a fixed
    # indent and --line does't display them. -- vila 10019

    def setUp(self):
        super().setUp()
        self.make_branch_with_diffs()

    def make_branch_with_diffs(self):
        level0 = self.make_branch_and_tree('level0')
        self.build_tree(['level0/file1', 'level0/file2'])
        level0.add('file1')
        level0.add('file2')
        self.wt_commit(level0, 'in branch level0')

        level1 = level0.controldir.sprout('level1').open_workingtree()
        self.build_tree_contents([('level1/file2', b'hello\n')])
        self.wt_commit(level1, 'in branch level1')
        level0.merge_from_branch(level1.branch)
        self.wt_commit(level0, 'merge branch level1')

    def _diff_file1_revno1(self):
        return b"""=== added file 'file1'
--- file1\t1970-01-01 00:00:00 +0000
+++ file1\t2005-11-22 00:00:00 +0000
@@ -0,0 +1,1 @@
+contents of level0/file1

"""

    def _diff_file2_revno2(self):
        return b"""=== modified file 'file2'
--- file2\t2005-11-22 00:00:00 +0000
+++ file2\t2005-11-22 00:00:01 +0000
@@ -1,1 +1,1 @@
-contents of level0/file2
+hello

"""

    def _diff_file2_revno1_1_1(self):
        return b"""=== modified file 'file2'
--- file2\t2005-11-22 00:00:00 +0000
+++ file2\t2005-11-22 00:00:01 +0000
@@ -1,1 +1,1 @@
-contents of level0/file2
+hello

"""

    def _diff_file2_revno1(self):
        return b"""=== added file 'file2'
--- file2\t1970-01-01 00:00:00 +0000
+++ file2\t2005-11-22 00:00:00 +0000
@@ -0,0 +1,1 @@
+contents of level0/file2

"""

    def assertLogRevnosAndDiff(self, args, expected,
                               working_dir='.'):
        self.run_bzr(['log', '-p'] + args, working_dir=working_dir)
        expected_revnos_and_depths = [
            (revno, depth) for revno, depth, diff in expected]
        # Check the revnos and depths first to make debugging easier
        self.assertEqual(expected_revnos_and_depths,
                         [(r.revno, r.merge_depth)
                          for r in self.get_captured_revisions()])
        # Now check the diffs, adding the revno  in case of failure
        fmt = 'In revno %s\n%s'
        for expected_rev, actual_rev in zip(expected,
                                            self.get_captured_revisions()):
            revno, depth, expected_diff = expected_rev
            actual_diff = actual_rev.diff
            self.assertEqualDiff(fmt % (revno, expected_diff),
                                 fmt % (revno, actual_diff))

    def test_log_diff_with_merges(self):
        self.assertLogRevnosAndDiff(
            ['-n0'],
            [('2', 0, self._diff_file2_revno2()),
             ('1.1.1', 1, self._diff_file2_revno1_1_1()),
             ('1', 0, self._diff_file1_revno1() +
              self._diff_file2_revno1())],
            working_dir='level0')

    def test_log_diff_file1(self):
        self.assertLogRevnosAndDiff(['-n0', 'file1'],
                                    [('1', 0, self._diff_file1_revno1())],
                                    working_dir='level0')

    def test_log_diff_file2(self):
        self.assertLogRevnosAndDiff(['-n1', 'file2'],
                                    [('2', 0, self._diff_file2_revno2()),
                                     ('1', 0, self._diff_file2_revno1())],
                                    working_dir='level0')


class TestLogUnicodeDiff(TestLog):

    def test_log_show_diff_non_ascii(self):
        # Smoke test for bug #328007 UnicodeDecodeError on 'log -p'
        message = 'Message with \xb5'
        body = b'Body with \xb5\n'
        wt = self.make_branch_and_tree('.')
        self.build_tree_contents([('foo', body)])
        wt.add('foo')
        wt.commit(message=message)
        # check that command won't fail with unicode error
        # don't care about exact output because we have other tests for this
        out, err = self.run_bzr('log -p --long')
        self.assertNotEqual('', out)
        self.assertEqual('', err)
        out, err = self.run_bzr('log -p --short')
        self.assertNotEqual('', out)
        self.assertEqual('', err)
        out, err = self.run_bzr('log -p --line')
        self.assertNotEqual('', out)
        self.assertEqual('', err)


class TestLogEncodings(tests.TestCaseInTempDir):

    _mu = '\xb5'
    _message = 'Message with \xb5'

    # Encodings which can encode mu
    good_encodings = [
        'utf-8',
        'latin-1',
        'iso-8859-1',
        'cp437',  # Common windows encoding
        'cp1251',  # Russian windows encoding
        'cp1258',  # Common windows encoding
    ]
    # Encodings which cannot encode mu
    bad_encodings = [
        'ascii',
        'iso-8859-2',
        'koi8_r',
    ]

    def setUp(self):
        super().setUp()
        self.overrideAttr(osutils, '_cached_user_encoding')

    def create_branch(self):
        brz = self.run_bzr
        brz('init')
        self.build_tree_contents([('a', b'some stuff\n')])
        brz('add a')
        brz(['commit', '-m', self._message])

    def try_encoding(self, encoding, fail=False):
        brz = self.run_bzr
        if fail:
            self.assertRaises(UnicodeEncodeError,
                              self._mu.encode, encoding)
            encoded_msg = self._message.encode(encoding, 'replace')
        else:
            encoded_msg = self._message.encode(encoding)

        old_encoding = osutils._cached_user_encoding
        # This test requires that 'run_bzr' uses the current
        # breezy, because we override user_encoding, and expect
        # it to be used
        try:
            osutils._cached_user_encoding = 'ascii'
            # We should be able to handle any encoding
            out, err = brz('log', encoding=encoding)
            if not fail:
                # Make sure we wrote mu as we expected it to exist
                self.assertNotEqual(-1, out.find(self._message))
            else:
                self.assertNotEqual(-1, out.find('Message with ?'))
        finally:
            osutils._cached_user_encoding = old_encoding

    def test_log_handles_encoding(self):
        self.create_branch()

        for encoding in self.good_encodings:
            self.try_encoding(encoding)

    def test_log_handles_bad_encoding(self):
        self.create_branch()

        for encoding in self.bad_encodings:
            self.try_encoding(encoding, fail=True)

    def test_stdout_encoding(self):
        brz = self.run_bzr
        osutils._cached_user_encoding = "cp1251"

        brz('init')
        self.build_tree(['a'])
        brz('add a')
        brz(['commit', '-m', '\u0422\u0435\u0441\u0442'])
        stdout, stderr = self.run_bzr_raw('log', encoding='cp866')

        message = stdout.splitlines()[-1]

        # explanation of the check:
        # u'\u0422\u0435\u0441\u0442' is word 'Test' in russian
        # in cp866  encoding this is string '\x92\xa5\xe1\xe2'
        # in cp1251 encoding this is string '\xd2\xe5\xf1\xf2'
        # This test should check that output of log command
        # encoded to sys.stdout.encoding
        test_in_cp866 = b'\x92\xa5\xe1\xe2'
        test_in_cp1251 = b'\xd2\xe5\xf1\xf2'
        # Make sure the log string is encoded in cp866
        self.assertEqual(test_in_cp866, message[2:])
        # Make sure the cp1251 string is not found anywhere
        self.assertEqual(-1, stdout.find(test_in_cp1251))


class TestLogFile(TestLogWithLogCatcher):

    def test_log_local_branch_file(self):
        """We should be able to log files in local treeless branches"""
        tree = self.make_branch_and_tree('tree')
        self.build_tree(['tree/file'])
        tree.add('file')
        tree.commit('revision 1')
        tree.controldir.destroy_workingtree()
        self.run_bzr('log tree/file')

    def prepare_tree(self, complex=False):
        # The complex configuration includes deletes and renames
        tree = self.make_branch_and_tree('parent')
        self.build_tree(['parent/file1', 'parent/file2', 'parent/file3'])
        tree.add('file1')
        tree.commit('add file1')
        tree.add('file2')
        tree.commit('add file2')
        tree.add('file3')
        tree.commit('add file3')
        child_tree = tree.controldir.sprout('child').open_workingtree()
        self.build_tree_contents([('child/file2', b'hello')])
        child_tree.commit(message='branch 1')
        tree.merge_from_branch(child_tree.branch)
        tree.commit(message='merge child branch')
        if complex:
            tree.remove('file2')
            tree.commit('remove file2')
            tree.rename_one('file3', 'file4')
            tree.commit('file3 is now called file4')
            tree.remove('file1')
            tree.commit('remove file1')
        os.chdir('parent')

    # FIXME: It would be good to parametrize the following tests against all
    # formatters. But the revisions selection is not *currently* part of the
    # LogFormatter contract, so using LogCatcher is sufficient -- vila 100118
    def test_log_file1(self):
        self.prepare_tree()
        self.assertLogRevnos(['-n0', 'file1'], ['1'])

    def test_log_file2(self):
        self.prepare_tree()
        # file2 full history
        self.assertLogRevnos(['-n0', 'file2'], ['4', '3.1.1', '2'])
        # file2 in a merge revision
        self.assertLogRevnos(['-n0', '-r3.1.1', 'file2'], ['3.1.1'])
        # file2 in a mainline revision
        self.assertLogRevnos(['-n0', '-r4', 'file2'], ['4', '3.1.1'])
        # file2 since a revision
        self.assertLogRevnos(['-n0', '-r3..', 'file2'], ['4', '3.1.1'])
        # file2 up to a revision
        self.assertLogRevnos(['-n0', '-r..3', 'file2'], ['2'])

    def test_log_file3(self):
        self.prepare_tree()
        self.assertLogRevnos(['-n0', 'file3'], ['3'])

    def test_log_file_historical_missing(self):
        # Check logging a deleted file gives an error if the
        # file isn't found at the end or start of the revision range
        self.prepare_tree(complex=True)
        err_msg = "Path unknown at end or start of revision range: file2"
        err = self.run_bzr('log file2', retcode=3)[1]
        self.assertContainsRe(err, err_msg)

    def test_log_file_historical_end(self):
        # Check logging a deleted file is ok if the file existed
        # at the end the revision range
        self.prepare_tree(complex=True)
        self.assertLogRevnos(['-n0', '-r..4', 'file2'], ['4', '3.1.1', '2'])

    def test_log_file_historical_start(self):
        # Check logging a deleted file is ok if the file existed
        # at the start of the revision range
        self.prepare_tree(complex=True)
        self.assertLogRevnos(['file1'], [])

    def test_log_file_renamed(self):
        """File matched against revision range, not current tree."""
        self.prepare_tree(complex=True)

        # Check logging a renamed file gives an error by default
        err_msg = "Path unknown at end or start of revision range: file3"
        err = self.run_bzr('log file3', retcode=3)[1]
        self.assertContainsRe(err, err_msg)

        # Check we can see a renamed file if we give the right end revision
        self.assertLogRevnos(['-r..4', 'file3'], ['3'])


class TestLogMultiple(TestLogWithLogCatcher):

    def prepare_tree(self):
        tree = self.make_branch_and_tree('parent')
        self.build_tree([
            'parent/file1',
            'parent/file2',
            'parent/dir1/',
            'parent/dir1/file5',
            'parent/dir1/dir2/',
            'parent/dir1/dir2/file3',
            'parent/file4'])
        tree.add('file1')
        tree.commit('add file1')
        tree.add('file2')
        tree.commit('add file2')
        tree.add(['dir1', 'dir1/dir2', 'dir1/dir2/file3'])
        tree.commit('add file3')
        tree.add('file4')
        tree.commit('add file4')
        tree.add('dir1/file5')
        tree.commit('add file5')
        child_tree = tree.controldir.sprout('child').open_workingtree()
        self.build_tree_contents([('child/file2', b'hello')])
        child_tree.commit(message='branch 1')
        tree.merge_from_branch(child_tree.branch)
        tree.commit(message='merge child branch')
        os.chdir('parent')

    def test_log_files(self):
        """The log for multiple file should only list revs for those files"""
        self.prepare_tree()
        self.assertLogRevnos(['file1', 'file2', 'dir1/dir2/file3'],
                             ['6', '5.1.1', '3', '2', '1'])

    def test_log_directory(self):
        """The log for a directory should show all nested files."""
        self.prepare_tree()
        self.assertLogRevnos(['dir1'], ['5', '3'])

    def test_log_nested_directory(self):
        """The log for a directory should show all nested files."""
        self.prepare_tree()
        self.assertLogRevnos(['dir1/dir2'], ['3'])

    def test_log_in_nested_directory(self):
        """The log for a directory should show all nested files."""
        self.prepare_tree()
        os.chdir("dir1")
        self.assertLogRevnos(['.'], ['5', '3'])

    def test_log_files_and_directories(self):
        """Logging files and directories together should be fine."""
        self.prepare_tree()
        self.assertLogRevnos(['file4', 'dir1/dir2'], ['4', '3'])

    def test_log_files_and_dirs_in_nested_directory(self):
        """The log for a directory should show all nested files."""
        self.prepare_tree()
        os.chdir("dir1")
        self.assertLogRevnos(['dir2', 'file5'], ['5', '3'])


class MainlineGhostTests(TestLogWithLogCatcher):

    def setUp(self):
        super().setUp()
        tree = self.make_branch_and_tree('')
        tree.set_parent_ids([b"spooky"], allow_leftmost_as_ghost=True)
        tree.add('')
        tree.commit('msg1', rev_id=b'rev1')
        tree.commit('msg2', rev_id=b'rev2')

    def test_log_range(self):
        self.assertLogRevnos(["-r1..2"], ["2", "1"])

    def test_log_norange(self):
        self.assertLogRevnos([], ["2", "1"])

    def test_log_range_open_begin(self):
        (stdout, stderr) = self.run_bzr(['log', '-r..2'], retcode=3)
        self.assertEqual(["2", "1"],
                         [r.revno for r in self.get_captured_revisions()])
        self.assertEqual("brz: ERROR: Further revision history missing.\n",
                         stderr)

    def test_log_range_open_end(self):
        self.assertLogRevnos(["-r1.."], ["2", "1"])


class TestLogMatch(TestLogWithLogCatcher):

    def prepare_tree(self):
        tree = self.make_branch_and_tree('')
        self.build_tree(
            ['/hello.txt', '/goodbye.txt'])
        tree.add('hello.txt')
        tree.commit(message='message1', committer='committer1',
                    authors=['author1'])
        tree.add('goodbye.txt')
        tree.commit(message='message2', committer='committer2',
                    authors=['author2'])

    def test_message(self):
        self.prepare_tree()
        self.assertLogRevnos(["-m", "message1"], ["1"])
        self.assertLogRevnos(["-m", "message2"], ["2"])
        self.assertLogRevnos(["-m", "message"], ["2", "1"])
        self.assertLogRevnos(["-m", "message1", "-m", "message2"], ["2", "1"])
        self.assertLogRevnos(["--match-message", "message1"], ["1"])
        self.assertLogRevnos(["--match-message", "message2"], ["2"])
        self.assertLogRevnos(["--match-message", "message"], ["2", "1"])
        self.assertLogRevnos(["--match-message", "message1",
                              "--match-message", "message2"], ["2", "1"])
        self.assertLogRevnos(["--message", "message1"], ["1"])
        self.assertLogRevnos(["--message", "message2"], ["2"])
        self.assertLogRevnos(["--message", "message"], ["2", "1"])
        self.assertLogRevnos(["--match-message", "message1",
                              "--message", "message2"], ["2", "1"])
        self.assertLogRevnos(["--message", "message1",
                              "--match-message", "message2"], ["2", "1"])

    def test_committer(self):
        self.prepare_tree()
        self.assertLogRevnos(["-m", "committer1"], ["1"])
        self.assertLogRevnos(["-m", "committer2"], ["2"])
        self.assertLogRevnos(["-m", "committer"], ["2", "1"])
        self.assertLogRevnos(["-m", "committer1", "-m", "committer2"],
                             ["2", "1"])
        self.assertLogRevnos(["--match-committer", "committer1"], ["1"])
        self.assertLogRevnos(["--match-committer", "committer2"], ["2"])
        self.assertLogRevnos(["--match-committer", "committer"], ["2", "1"])
        self.assertLogRevnos(["--match-committer", "committer1",
                              "--match-committer", "committer2"], ["2", "1"])

    def test_author(self):
        self.prepare_tree()
        self.assertLogRevnos(["-m", "author1"], ["1"])
        self.assertLogRevnos(["-m", "author2"], ["2"])
        self.assertLogRevnos(["-m", "author"], ["2", "1"])
        self.assertLogRevnos(["-m", "author1", "-m", "author2"],
                             ["2", "1"])
        self.assertLogRevnos(["--match-author", "author1"], ["1"])
        self.assertLogRevnos(["--match-author", "author2"], ["2"])
        self.assertLogRevnos(["--match-author", "author"], ["2", "1"])
        self.assertLogRevnos(["--match-author", "author1",
                              "--match-author", "author2"], ["2", "1"])



# Copyright (C) 2007-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for repository write groups."""

from breezy import errors
from breezy.tests import per_repository, test_server
from breezy.transport import memory


class TestWriteGroup(per_repository.TestCaseWithRepository):

    def test_start_write_group_unlocked_needs_write_lock(self):
        repo = self.make_repository('.')
        self.assertRaises(errors.NotWriteLocked, repo.start_write_group)

    def test_start_write_group_read_locked_needs_write_lock(self):
        repo = self.make_repository('.')
        with repo.lock_read():
            self.assertRaises(errors.NotWriteLocked, repo.start_write_group)

    def test_start_write_group_write_locked_gets_None(self):
        repo = self.make_repository('.')
        repo.lock_write()
        self.assertEqual(None, repo.start_write_group())
        repo.commit_write_group()
        repo.unlock()

    def test_start_write_group_twice_errors(self):
        repo = self.make_repository('.')
        repo.lock_write()
        repo.start_write_group()
        try:
            # don't need a specific exception for now - this is
            # really to be sure it's used right, not for signalling
            # semantic information.
            self.assertRaises(errors.BzrError, repo.start_write_group)
        finally:
            repo.commit_write_group()
            repo.unlock()

    def test_commit_write_group_does_not_error(self):
        repo = self.make_repository('.')
        repo.lock_write()
        repo.start_write_group()
        # commit_write_group can either return None (for repositories without
        # isolated transactions) or a hint for pack(). So we only check it
        # works in this interface test, because all repositories are exercised.
        repo.commit_write_group()
        repo.unlock()

    def test_unlock_in_write_group(self):
        repo = self.make_repository('.')
        repo.lock_write()
        repo.start_write_group()
        # don't need a specific exception for now - this is
        # really to be sure it's used right, not for signalling
        # semantic information.
        self.assertLogsError(errors.BzrError, repo.unlock)
        # after this error occurs, the repository is unlocked, and the write
        # group is gone.  you've had your chance, and you blew it. ;-)
        self.assertFalse(repo.is_locked())
        self.assertRaises(errors.BzrError, repo.commit_write_group)
        self.assertRaises(errors.BzrError, repo.unlock)

    def test_is_in_write_group(self):
        repo = self.make_repository('.')
        self.assertFalse(repo.is_in_write_group())
        repo.lock_write()
        repo.start_write_group()
        self.assertTrue(repo.is_in_write_group())
        repo.commit_write_group()
        self.assertFalse(repo.is_in_write_group())
        # abort also removes the in_write_group status.
        repo.start_write_group()
        self.assertTrue(repo.is_in_write_group())
        repo.abort_write_group()
        self.assertFalse(repo.is_in_write_group())
        repo.unlock()

    def test_abort_write_group_gets_None(self):
        repo = self.make_repository('.')
        repo.lock_write()
        repo.start_write_group()
        self.assertEqual(None, repo.abort_write_group())
        repo.unlock()

    def test_abort_write_group_does_not_raise_when_suppressed(self):
        if self.transport_server is test_server.LocalURLServer:
            self.transport_server = None
        self.vfs_transport_factory = memory.MemoryServer
        repo = self.make_repository('repo')
        token = repo.lock_write()
        self.addCleanup(repo.unlock)
        repo.start_write_group()
        # Damage the repository on the filesystem
        t = self.get_transport('')
        t.rename('repo', 'foo')
        self.addCleanup(t.rename, 'foo', 'repo')
        # abort_write_group will not raise an error, because either an
        # exception was not generated, or the exception was caught and
        # suppressed.  See also test_pack_repository's test of the same name.
        self.assertEqual(None, repo.abort_write_group(suppress_errors=True))



#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Flake8 type checking before commits.

Set the ``flake8.pre_commit_check`` configuration variable to True
to enable this.
"""


from breezy.errors import BzrError, DependencyNotPresent


class Flake8Errors(BzrError):

    _fmt = (
        "Running in strict flake8 mode; aborting commit, "
        "since %(errors)d flake8 errors exist."
    )

    def __init__(self, errors):
        self.errors = errors


def _delta_files(tree_delta):
    for path, file_id, kind in tree_delta.added:
        if kind == "file":
            yield path
    for (
        path,
        file_id,
        kind,
        text_modified,
        meta_modified,
    ) in tree_delta.modified:
        if kind == "file" and text_modified:
            yield path
    for (
        oldpath,
        newpath,
        id,
        kind,
        text_modified,
        meta_modified,
    ) in tree_delta.renamed:
        if kind == "file" and text_modified:
            yield newpath
    for path, id, old_kind, new_kind in tree_delta.kind_changed:
        if new_kind == "file":
            yield path


def _copy_files_to(tree, target_dir, files):
    from breezy import osutils

    for relpath in files:
        target_path = os.path.join(target_dir, relpath)
        os.makedirs(os.path.dirname(target_path))
        with tree.get_file(relpath) as inf, open(target_path, "wb") as outf:
            osutils.pumpfile(inf, outf)
            yield target_path


def _update_paths(checker_manager, temp_prefix):
    temp_prefix_length = len(temp_prefix)
    for checker in checker_manager.checkers:
        filename = checker.display_name
        if filename.startswith(temp_prefix):
            checker.display_name = os.path.relpath(
                filename[temp_prefix_length:]
            )


def hook(config, tree_delta, future_tree):
    """Execute Flake8 on the files in git's index.

    Determine which files are about to be committed and run Flake8 over
    them to check for violations.

    :param bool strict:
        If True, return the total number of errors/violations found by
        Flake8. This will cause the hook to fail.
    :returns:
        Total number of errors found during the run.
    :rtype:
        int
    """
    try:
        from flake8.main import application
    except ModuleNotFoundError as e:
        raise DependencyNotPresent('flake8', e)
    import tempfile

    strict = config.get("flake8.strict")

    app = application.Application()
    with tempfile.TemporaryDirectory() as tempdir:
        filepaths = list(
            _copy_files_to(future_tree, tempdir, _delta_files(tree_delta))
        )
        app.initialize([tempdir])
        app.options._running_from_vcs = True
        app.run_checks(filepaths)
        _update_paths(app.file_checker_manager, tempdir)
        app.report_errors()

    if strict:
        if app.result_count:
            raise Flake8Errors(app.result_count)


def _check_flake8(local, master, old_revno, old_revid, future_revno,
                  future_revid, tree_delta, future_tree):
    branch = local or master
    config = branch.get_config_stack()
    if config.get("flake8.pre_commit_check"):
        hook(config, tree_delta, future_tree)


from breezy.branch import Branch

Branch.hooks.install_named_hook("pre_commit", _check_flake8, "Check flake8")  # type: ignore
